datastore_search = '''
This action allows you to search data in a resource. It is designed to function in a similar way
to CKAN's core datastore_search but with a few extra bells and whistles, most prominently
versioning. This allows the resource to be searched at any moment in it's lifespan and have the
data as it looked at that moment returned, even if it has changed since.

If the resource to be searched is private then appropriate authorization is required.

Note that in the parameters listed below spaces should only be included if part of a field name,
so, for example, don't include any spaces in comma separated lists unless needed.

:param resource_id: id of the resource to be searched against
:type resource_id: string
:param q: full text query. If a string is passed, all fields are searched with the value. If a
      dict is passed each of the fields and values contained within will be searched as
      required (e.g. {"field1": "a", "field2": "b"}).
:type q: string or dictionary
:param filters: a dictionary of conditions that must be met to match a record
                (e.g {"field1": "a", "field2": "b"}) (optional)
:type filters: dictionary
:param after: search_after value for elasticsearch to paginate from (optional). Use this
              mechanism to do deep (beyond 10000 values) pagination. The values have to match
              the sort currently in use and therefore it's recommended that this value is not
              built but rather passed from the previous result's 'after' key.
:type after: a list of values
:param limit: maximum number of records to return (optional, default: 100)
:type limit: int
:param offset: offset this number of records (optional)
:type offset: int
:param fields: fields to return for each record (optional, default: all fields are returned)
:type fields: list or comma separated string
:param sort: list of field names with ordering. Ordering is ascending by default, if descending
             is required, add "desc" after the field name
             e.g.: "fieldname1,fieldname2 desc" sorts by fieldname1 asc and fieldname2 desc
:type sort: list or comma separated string
:param version: version to search at, if not provided the current version of the data is
               searched.
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param facets: if present, the top 10 most frequent values for each of the fields in this list
               will be returned along with estimated counts for each value. Calculating these
               results has a reasonable overhead so only include this parameter if you need it
:type facets: list or comma separated string
:param facet_limits: if present, specifies the number of top values to retrieve for the facets
                    listed within. The default number will be used if this parameter is not
                    specified or if a facet in the facets list does not appear in this dict. For
                    example, with this facet list ['facet1', 'facet2', 'facet3', 'facet4'], and
                    this facet_limits dict: {'facet1': 50, 'facet4': 10}, facet1 and facet4
                    would be limited to top 50 and 10 values respectively, whereas facet2 and
                    facet3 would be limited to the default of the top 10.
:type facet_limits: a dict
:param run_query: boolean value indicating whether the query should be run and the results
                  returned or whether the query should be created and the elasticsearch query
                  returned instead of the results. Defaults to True.
:type run_query: boolean


**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param fields: fields/columns and their extra metadata
:type fields: list of dicts
:param total: number of total matching records
:type total: int
:param records: list of matching results
:type records: list of dicts
:param facets: list of fields and their top 10 values, if requested
:type facets: dict
:param after: the next page's search_after value which can be passed back as the "after"
              parameter. This value will always be included if there were results otherwise None
              is returned. A value will also always be returned even if this page is the last.
:type after: a list or None

If run_query is True, then a dict with the following keys is returned instead:

:param indexes: a list of the fully qualified indexes that the query would have been run against
:type indexes: a list of strings
:param search: the query dict that would have been sent to elasticsearch
:type search: dict

In addition to returning these result dicts, the actual result object is made available through
the context dict under the key "versioned_datastore_query_result". This isn't available through
the http action API however.
'''

datastore_create = '''
Adds a resource to the versioned datastore. This action doesn't take any data, it simply ensures
any setup work is complete for the given resource in the search backend. To add data after
creating a resource in the datastore, use the datastore_upsert action.

:param resource_id: resource id of the resource
:type resource_id: string

**Results:**

:returns: True if the datastore was initialised for this resource (or indeed if it was already
          initialised) and False if not. If False is returned this implies that the resource
          cannot be ingested into the datastore because the format is not supported
:rtype: boolean
'''

datastore_upsert = '''
Upserts data into the datastore for the resource. The data can be provided in the data_dict
using the key 'records' or, if data is not specified, the URL on the resource is used.

:param resource_id: resource id of the resource
:type resource_id: string
:param replace: whether to remove any records not included in this update. If True, any record
                that was not in the set of data ingested is marked as deleted. Note that this is
                any record not included, not any record that hasn't changed. If a record is
                included in the data, has an _id that matches an existing record and its data
                hasn't changed, it will not be removed.
                If False, the data is ingested and all existing data is left to be either
                updated or continue to be current. There is no default for this option.
:type replace: bool
:param version: the version to store the data under (optional, if not specified defaults to now)
:type version: int

**Results:**

:returns: details about the job that has been submitted to fulfill the upsert request.
:rtype: dict
'''

datastore_delete = '''
Deletes the data in the datastore against the given resource_id. Note that this is achieved by
setting all records to be empty in a new version and then indexing that new version. This
ensures that the data is not available in the latest version but is in old ones.

:param resource_id: resource id of the resource
:type resource_id: string
:param version: the version to delete the data at, can be missing and if it is it's defaults to
                the current timestamp
:type version: integer
'''

datastore_reindex = '''
Triggers a reindex of the given resource's data. This does not reingest the data to mongo, but
it does reindex the data in mongo to elasticsearch. The intent of this action is to allow
mapping changes (for example) to be picked up.

Data dict params:
:param resource_id: resource id that the record id appears in
:type resource_id: string

**Results:**

:returns: a dict containing the details of the reindex as returned from elasticsearch
:rtype: dict
'''

datastore_ensure_privacy = '''
Ensure that the privacy settings are correct across all resources in the datastore or for just
one resource.

Params:
:param resource_id: optionally, the id of a specific resource to update. If this is present then
                    only the resource provided will have it's privacy updated. If it is not
                    present, all resources are updated.
:type resource_id: string


**Results:**
The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param modified: the number of resources that had their privacy setting modified
:type ensured: integer
:param total: the total number of resources examined
:type total: integer
'''

datastore_autocomplete = '''
Provides autocompletion results against a specific field in a specific resource.

**Data dict params:**

:param resource_id: id of the resource to be searched against
:type resource_id: string
:param q: full text query. If a string is passed, all fields are searched with the value. If a
      dict is passed each of the fields and values contained within will be searched as
      required (e.g. {"field1": "a", "field2": "b"}).
:type q: string or dictionary
:param filters: a dictionary of conditions that must be met to match a record
                (e.g {"field1": "a", "field2": "b"}) (optional)
:type filters: dictionary
:param limit: maximum number of records to return (optional, default: 100)
:type limit: int
:param after: search after offset value as a base64 encoded string
:type after: string
:param field: the field to autocomplete against
:type field: string
:param term: the search term for the autocompletion
:type term: string
:param version: version to search at, if not provided the current version of the data is
               searched.
:type version: int, number of milliseconds (not seconds!) since UNIX epoch


**Results:**

:returns: a dict containing the list of values and an after value for the next page's results
:rtype: dict
'''

datastore_query_extent = '''
Return the geospatial extent of the results of a given datastore search query. The data_dict
parameters are the same as the arguments for `datastore_search`.


**Results:**

:rtype: A dict with the following keys
:param total_count: total number of rows matching the query
:type total_count: integer
:param geom_count: Number of rows matching the query that have geospatial information
:type geom_count: int
:param bounds: the extent of the query's results, this will be missing if no bound can be
               calculated (for example, if the resource has no geo data)
:type bounds: list in the format [[lat min, long min], [lat max, long max]]
'''

datastore_search_raw = '''
This action allows you to search data in a resource using a raw elasticsearch query. This action
allows more flexibility over the search both in terms of querying using any of elasticsearch's
different DSL queries as well as aspects like turning versioning on and off.

If the resource to be searched is private then appropriate authorization is required.

Note that the structure of the documents in elasticsearch is defined as such:

    - data._id: the id of the record, this is always an integer
    - data.*: the data fields. Each field is stored in 3 different ways:
        - data.<field_name>: keyword type
        - data.<field_name>.full: text type
        - data.<field_name>.number: double type, will be missing if the data value isn't
                                    convertable to a number
    - meta.*: various metadata fields, including:
        - meta.all: a text type field populated from all the data in the data.* fields
        - meta.geo: if a pair of lat/lon fields have been assigned on this resource this geo
                    point type field is available
        - meta.version: the version of this record, this field is a date type field represented
                        in epoch millis
        - meta.next_version: the next version of this record, this field is a date type field
                        represented in epoch millis. If missing this is the current version of
                        the record
        - meta.versions: a date range type field which encapsulates the version range this
                         document applies to for the record

Params:

:param resource_id: id of the resource to be searched against, required
:type resource_id: string
:param search: the search JSON to submit to elasticsearch. This should be a valid elasticsearch
               search, the only modifications that will be made to it are setting the version
               filter unless include_version=False. If not included then an empty {} is used.
:type search: dict
:param version: version to search at, if not provided the current version of the data is
               searched (unless using include_version=False).
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param raw_result: whether to parse the result and return in the same format as datastore_search
                   or just return the exact elasticsearch response, unaltered. Defaults to False
:type raw_result: bool
:param include_version: whether to include the version filter or not. By default this is True.
                        This can only be set to False in combination with raw_result=True.
:type include_version: bool


**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param fields: fields/columns and their extra metadata
:type fields: list of dicts
:param total: number of total matching records
:type total: int
:param records: list of matching results
:type records: list of dicts
:param facets: list of fields and their top 10 values, if requested
:type facets: dict
:param after: the next page's search_after value which can be passed back as the "after"
              parameter. This value will always be included if there were results otherwise None
              is returned. A value will also always be returned even if this page is the last.
:type after: a list or None

In addition to returning this result, the actual result object is made available through the
context dict under the key "versioned_datastore_query_result". This isn't available through the
http action API however.

If raw_result is True, then the elasticsearch response is returned without modification.
'''

datastore_get_record_versions = '''
Given a record id and an resource it appears in, returns the version timestamps available for
that record in ascending order.

Data dict params:
:param resource_id: resource id that the record id appears in
:type resource_id: string
:param id: the id of the record
:type id: integer

**Results:**

:returns: a list of versions
:rtype: list
'''

datastore_get_resource_versions = '''
Given a resource id, returns the version timestamps available for that resource in ascending
order along with the number of records modified in the version and the number of records at that
version.

This action also accepts all the datastore_search parameters that can be used to search the
resource's data (i.e. q, filters). If these parameters are passed then the returned versions and
counts are for the data found using the search. This allows the discovery of the versions and
counts available for a query's result set.

Data dict params:
:param resource_id: resource id
:type resource_id: string

**Results:**

:returns: a list of dicts, each in the form: {"version": #, "changes": #, "count": #}
:rtype: list of dicts
'''

datastore_get_rounded_version = '''
Round the requested version of this query down to the nearest actual version of the
resource. This is necessary because we work in a system where although you can just query at
a timestamp you should round it down to the nearest known version. This guarantees that when
you come back and search the data later yo'll get the same data. If the version requested
is older than the oldest version of the resource then the requested version itself is
returned (this is just a choice I made, we could return 0 for example instead).

An example: a version of resource A is created at t=2 and then a search is completed on it
at t=5. If a new version is created at t=3 then the search at t=5 won't return the data it
should. We could solve this problem in one of two ways:

- limit new versions of resources to only be either older than the oldest saved search
- rely on the current time when resource versions are created and searches are saved

There are however issues with both of these approaches:

- forcing new resource versions to exceed currently created search versions would
  require a reasonable amount of work to figure out what the latest search version is and
  also crosses extension boundaries as we'd need access to any tables that have a latest
  query version.
- we want to be able to create resource versions beyond the latest version in the system
  but before the current time to accommodate non-live data (i.e. data that comes from
  timestamped dumps). There are a few benefits to allowing this, for example it allows
  loading data where we create a number of resource versions at the same time but the
  versions themselves represent when the data was extacted from another system or indeed
  created rather than when it was loaded into CKAN.

Data dict params:
:param resource_id: the resource id
:type resource_id: string
:param version: the version to round (optional, if missing the latest version is returned)
:type version: integer

**Results:**

:returns: the rounded version or None if no versions are available for the given resource id
:rtype: integer or None
'''

datastore_multisearch = '''
This action allows you to search data in multiple resources.

The resources that are searched for the in this action and the version they are searched at are
both extracted from the resource_ids_and_versions in the first instance, and if no information
in there is found then the resource_ids and version parameters are used as fall backs.
Regardless of where the resource ids list comes from though, it is always checked against
permissions to ensure the user has the right to access the given resources. Any resources
included that the user doesn't have access to are not searched and will be returned in the
"skipped_resources" part of the return value.


Params:

:param query: the search JSON
:type query: dict
:param version: version to search at, if not provided the current version of the data is
               searched
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param resource_ids_and_versions: a dict of resource ids and the versions to search them at. If
                                  this is present it's values are prioritised over the version
                                  and resource_ids parameters.
:type resource_ids_and_versions: dict of strings -> ints (number of milliseconds (not seconds!)
                                 since UNIX epoch)
:param query_version: the query language version (for example v1.0.0)
:type query_version: string
:param resource_ids: a list of resource ids to search. If no resources ids are specified (either
                     because the parameter is missing or because an empty list is passed) then
                     all resources in the datastore that the user can access are searched. Any
                     resources that the user cannot access or that aren't datastore resources
                     are skipped. If this means that no resources are available from the
                     provided list then a ValidationError is raised.
:type resource_ids: a list of strings
:param after: provides pagination. By passing a previous result set's after value, the next
              page's results can be found. If not provided then the first page is retrieved
:type after: a list
:param size: the number of records to return in the search result. If not provided then the
             default value of 100 is used. This value must be between 0 and 1000 and will be
             capped at which ever end is necessary if it is beyond these bounds
:type size: int
:param top_resources: whether to include the top 10 resources in the query result, default False
:type top_resources: bool
:param timings: whether to include timings in the result dict, default False. This is intended for
                developer debugging.
:type timings: bool

**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param total: number of total matching records
:type total: int
:param records: list of matching results as dicts. Each dict will contain a "data" key which
                holds the record data, and a "resource" key which holds the resource id the
                record belongs to.
:type records: list of dicts
:param after: the next page's search_after value which can be passed back in the "after"
              parameter. This value will always be included if there were results otherwise None
              is returned. A value will also always be returned even if this page is the last.
:type after: a list or None
:param top_resources: if requested, the top 10 resources and the number of records matched in
                      them by the query
:type top_resources: list of dicts
:param skipped_resources: a list of the resources from the requested resource_ids parameter that
                          were not searched. This list is only populated if a the resource ids
                          paramater is passed, otherwise it will be an empty list. The resources
                          in this list will have been skipped for one of two reasons: either the
                          user doesn't have access to the requested resource or the requested
                          resource isn't a datastore resource.
:type skipped_resources: a list of strings
:param timings: dict of events and how long they took as part of the response processing. This is
                only included in the response if the timings parameter is True
:type timings: dict
'''

datastore_create_slug = '''
Create a query slug based on the provided query parameters.

This action returns a slug which can be used to retrieve the query parameters passed (not the
query results) at a later time. This slug will return, if resolved using the
datastore_resolve_slug action, the query, the query version, the resource ids, the version and
the specific resources/versions map passed in to this action to create the slug. A few important
notes:

    - the slug returned will always be the same for the same search parameter combinations

    - the slugs never expire

    - if the version parameter is missing or null/None, the resolved slug will reflect this and
      not provide a version value - essentially we don't default it to the current time if it's
      missing


Params:

:param query: the search JSON
:type query: dict
:param version: version to search at, if this value is null or missing then the slug will not
                include version information
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param query_version: the query language version (for example v1.0.0)
:type query_version: string
:param resource_ids_and_versions: a dict of resource ids and the versions to search them at. If
                                  this is present it's values are prioritised over the version
                                  and resource_ids parameters.
:type resource_ids_and_versions: dict of strings -> ints (number of milliseconds (not seconds!)
                                 since UNIX epoch)
:param resource_ids: a list of resource ids to search
:type resource_ids: a list of strings
:param pretty_slug: whether to return a pretty slug with words or a uuid. Default: True
:type pretty_slug: bool

**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param slug: the slug
:type slug: string
:param is_new: whether the returned slug was newly created or already existed
:type is_new: bool
'''

datastore_resolve_slug = '''
Given a slug, resolves it and returns the query information associated with it.

Params:

:param slug: the slug to resolve
:type slug: string

**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param query: the query body
:type query: dict
:param query_version: the query version in use
:type query_version: string
:param version: the version the query is using
:type version: integer
:param resource_ids: the resource ids under search in the query
:type resource_ids: a list of strings
:param resource_ids_and_versions: a dict of resource ids -> versions to search them at
:type resource_ids_and_versions: a dict
:param created: the date time the slug was originally created
:type created: datetime in isoformat
'''

datastore_field_autocomplete = '''
Returns a dictionary of available fields in the datastore which contain the passed text. The
fields will be retrieved from resources available to the user. If a list of resource ids is
passed as a parameter then the resources from that list that the user has access to will be
used.

Params:

:param text: prefix to match the fields against. Optional, by default all fields are matched
:type text: string
:param resource_ids: the resources to find the fields in. Optional, by default all resources
                     available to the user are used
:type resource_ids: list of string resource ids, separated by commas
:param lowercase: whether to do a case insensitive prefix match. Optional, default: False
:type lowercase: bool

**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param fields: a dict of field names and their field properties. Example:
               {
                   "field1": {
                       "type": "keyword",
                       "fields": {
                           "full": "text",
                           "number": "double"
                       }
                   },
               }
               The type value is the index type of the field if accessed directly. The fields
               value indicates the presence of subfields with their names and associated index
               types included.
:type fields: dict
:param count: the number of fields returned
:type count: int
'''

datastore_value_autocomplete = '''
Finds values in the given field, from the given resources, which start with the given prefix and
returns up to {size} of them in a list. The values are sorted in alphabetical order.

Params:

:param field: the name of the field to autocomplete
:type field: str
:param prefix: the value to use as the prefix search, can be missing or blank
:type prefix: str
:param query: the search JSON
:type query: dict
:param version: version to search at, if not provided the current version of the data is
               searched
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param resource_ids_and_versions: a dict of resource ids and the versions to search them at. If
                                  this is present it's values are prioritised over the version
                                  and resource_ids parameters.
:type resource_ids_and_versions: dict of strings -> ints (number of milliseconds (not seconds!)
                                 since UNIX epoch)
:param query_version: the query language version (for example v1.0.0)
:type query_version: string
:param resource_ids: a list of resource ids to search. If no resources ids are specified (either
                     because the parameter is missing or because an empty list is passed) then
                     all resources in the datastore that the user can access are searched. Any
                     resources that the user cannot access or that aren't datastore resources
                     are skipped. If this means that no resources are available from the
                     provided list then a ValidationError is raised.
:type resource_ids: a list of strings
:param after: provides pagination. By passing a previous result set's after value, the next
              page's results can be found. If not provided then the first page is retrieved
:type after: a list
:param size: the number of records to return in the search result. If not provided then the
             default value of 20 is used. This value must be between 1 and 20 and will be capped at
             which ever end as necessary if it is beyond these bounds
:type size: int

**Results:**

The result of this action is a dictionary with the following keys:

:rtype: A dict with the following keys
:param total: number of total matching records
:type total: int
:param records: list of matching results as dicts. Each dict will contain a "data" key which
                holds the record data, and a "resource" key which holds the resource id the
                record belongs to.
:type records: list of dicts
:param after: the next page's search_after value which can be passed back in the "after"
              parameter. This value will always be included if there were results otherwise None
              is returned. A value will also always be returned even if this page is the last.
:type after: a list or None
:param top_resources: if requested, the top 10 resources and the number of records matched in
                      them by the query
:type top_resources: list of dicts
:param skipped_resources: a list of the resources from the requested resource_ids parameter that
                          were not searched. This list is only populated if a the resource ids
                          paramater is passed, otherwise it will be an empty list. The resources
                          in this list will have been skipped for one of two reasons: either the
                          user doesn't have access to the requested resource or the requested
                          resource isn't a datastore resource.
:type skipped_resources: a list of strings
:param timings: dict of events and how long they took as part of the response processing. This is
                only included in the response if the timings parameter is True
:type timings: dict

'''

datastore_queue_download = '''
Queues a task to generate a downloadable zip containing the data produced by the given query.

**Results:**

:returns: details about the job that has been submitted to fulfill the upsert request.
:rtype: dict
'''

datastore_regenerate_download = '''
Calls datastore_queue_download to regenerate a previous request. Note that notifier args
are still required as these are not stored in the original request, and server args may
be specified to override any stored ones.

**Results:**

:returns: details about the job that has been submitted to fulfill the upsert request.
:rtype: dict
'''

datastore_guess_fields = '''
This action allows you to retrieve a set of fields to display by default for a given search across
potentially multiple resources. The returned list of groups of fields is ordered by the number of
resources the fields in each group appears in under the provided query. Ties are handled by ordering
the group with the fields that appear in the most records first. If there is only one resource id
passed then the groups returned are ordered in ingestion order, if available.

The resources that are searched in this action and the version they are searched at are both
extracted from the resource_ids_and_versions in the first instance, and if no information in there
is found then the resource_ids and version parameters are used as fall backs. Regardless of where
the resource ids list comes from though, it is always checked against permissions to ensure the user
has the right to access the given resources.

Params:

:param query: the search JSON
:type query: dict
:param query_version: the query language version (for example v1.0.0)
:type query_version: string
:param version: version to search at, if not provided the current version of the data is searched
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param resource_ids_and_versions: a dict of resource ids and the versions to search them at. If this
                                  is present it's values are prioritised over the version and
                                  resource_ids parameters.
:type resource_ids_and_versions: dict of strings -> ints (number of milliseconds (not seconds!)
                                 since UNIX epoch)
:param resource_ids: a list of resource ids to search. If no resources ids are specified (either
                     because the parameter is missing or because an empty list is passed) then
                     all resources in the datastore that the user can access are searched. Any
                     resources that the user cannot access or that aren't datastore resources
                     are skipped. If this means that no resources are available from the
                     provided list then a ValidationError is raised.
:type resource_ids: a list of strings
:param size: the number of field groups to return in the search result. If not provided then the
             default is 10. Each group can have many fields in it from many resources. Each group is
             created by finding common fields across resources - fields are matched by comparing
             them case-insensitively.
:type size: int
:param ignore_groups: names of groups to ignore from the returned list. Note that this is groups,
                      not fields!
:type ignore_groups: a list of strings

**Results:**

The result of this action is a list of dicts, each with the following keys:

:rtype: A list of dicts with the following keys
:param group: the name of the group
:type group: string
:param count: the number of resources including fields in the group
:type count: int
:params records: the number of records the group's fields appear in
:type records: int
:param fields: a dict of field names -> list of resource ids representing the fields in the group
               and the resources they come from
:type fields: dict
'''

datastore_hash_query = '''
This action simply hashes the given query and returns the hex digest of it. The hash is produced
using sha1 and a custom algorithm - it's not just a hash of the query dict.

Params:

:param query: the search JSON
:type query: dict
:param query_version: the query language version (for example v1.0.0)
:type query_version: string

Returns:
:rtype: string
'''

datastore_is_datastore_resource = '''
This action checks whether the given resource is in the datastore.

Params:

:param resource_id: the resource id to check
:type resource_id: string

Returns:

:rtype: boolean
'''

datastore_get_latest_query_schema_version = '''
This action simply returns the latest available query schema version.

Returns:

:rtype: string
'''

datastore_count = '''
Count the number of records available at a specific version across a set of resources. This
allows quick counting of total records without any query, if you want to count with a query,
use the search actions with a limit of 0.

Params:

:param resource_ids: optionally, the ids of specific resources to count. Only public resources
                     can be counted. If this parameter is not provided, then all public
                     resources are counted.
:type resource_id: string, list of comma separated strings
:param version: version to count at, if not provided the current timestamp is used
:type version: int, number of milliseconds (not seconds!) since UNIX epoch

**Results:**

The result of this action is a dictionary with the following keys:
:rtype: an integer count
'''

datastore_edit_slug = '''
Add or modify the reserved slug for a query. Reserved slugs can only be replaced by sysadmins, but
if one has not been added yet for a query, any logged-in user can supply it.

Params:

:param current_slug: the slug currently referencing the search (reserved or non-reserved)
:type current_slug: str
:param new_reserved_slug: the string to use as the new reserved slug
:type new_reserved_slug: str

Returns:

:rtype: bool
'''

datastore_multisearch_counts = """
Count the number of records that match the query in each of the provided resources and
return the counts for each resource in a dict.

The query and resource IDs are parsed in the same way as the datastore_multisearch
action.

Params:

:param query: the search JSON
:type query: dict
:param version: version to search at, if not provided the current version of the data is
               searched
:type version: int, number of milliseconds (not seconds!) since UNIX epoch
:param resource_ids_and_versions: a dict of resource ids and the versions to search them
                                  at. If this is present it's values are prioritised
                                  over the version and resource_ids parameters.
:type resource_ids_and_versions: dict of strings -> ints (number of milliseconds (not
                                 seconds!) since UNIX epoch)
:param query_version: the query language version (for example v1.0.0)
:type query_version: string
:param resource_ids: a list of resource ids to search. If no resources ids are specified
                     (either because the parameter is missing or because an empty list
                     is passed) then all resources in the datastore that the user can
                     access are searched. Any resources that the user cannot access or
                     that aren't datastore resources are skipped. If this means that no
                     resources are available from the provided list then a
                     ValidationError is raised.
:type resource_ids: a list of strings

**Results:**

The result of this action is a dictionary where the keys are the resource IDs and the
values are the number of records in that resource which matched the query.

:rtype: dict with str keys and int values
"""



import json
from collections import OrderedDict

from ckan import model

from ...model.details import DatastoreResourceDetails


def create_details(resource_id, version, columns, file_hash=None):
    """
    Helper for creating a DatastoreResourceDetails object.

    :param resource_id: the resource id of the ingested resource
    :param version: the version being ingested
    :param columns: the columns present in the ingested resource
    :param file_hash: the computed file hash of the uploaded data
    :return: the id of the created row
    """
    columns_str = json.dumps(columns)
    details = DatastoreResourceDetails(
        resource_id=resource_id,
        version=version,
        columns=columns_str,
        file_hash=file_hash,
    )
    details.add()
    details.commit()
    return details.id


def get_details(resource_id, version):
    """
    Retrieve the details for a resource at a given version.

    :param resource_id: the resource id
    :param version: the version to get the details at
    :return: None or a DatastoreResourceDetails object
    """
    return (
        model.Session.query(DatastoreResourceDetails)
        .filter(DatastoreResourceDetails.resource_id == resource_id)
        .filter(DatastoreResourceDetails.version == version)
        .first()
    )


def get_all_details(resource_id, up_to_version=None):
    """
    Retrieve all the details for a resource and return them as an OrderedDict using the
    versions as the keys. If the up_to_version parameter is passed then any versions
    beyond it are ignored and not returned in the resulting OrderedDict.

    :param resource_id: the resource id
    :param up_to_version: the maximum version to include in the resulting OrderedDict (inclusive)
    :return: None or an OrderedDict of version: DatastoreResourceDetails objects in ascending order
    """
    query = (
        model.Session.query(DatastoreResourceDetails)
        .filter(DatastoreResourceDetails.resource_id == resource_id)
        .order_by(DatastoreResourceDetails.version.asc())
    )

    all_details = OrderedDict()
    for details in query:
        if up_to_version is not None and details.version > up_to_version:
            break
        else:
            all_details[details.version] = details

    return all_details


def get_last_file_hash(resource_id):
    """
    Retrieves the most recent file hash and returns it. If there is no most recent file
    hash (either because there isn't a file hash present or because there aren't any
    details rows available, then this function returns None.

    :param resource_id: the resource id
    :return: None or the most recent file hash
    """
    last_details = (
        model.Session.query(DatastoreResourceDetails)
        .filter(DatastoreResourceDetails.resource_id == resource_id)
        .order_by(DatastoreResourceDetails.version.desc())
        .first()
    )

    return last_details.file_hash if last_details is not None else None



import json
from ckan.plugins import toolkit
from datetime import date

from .lib.common import ALL_FORMATS
from .lib.importing import stats
from .lib.query.slugs import create_nav_slug


def is_duplicate_ingestion(stat):
    """
    Detects whether the error message on this ImportStats object is a
    DuplicateDataSource error message or not. This is based on the error containing the
    phrase "this file has been ingested before" and is therefore a quite fragile, but
    it's only for UI purposes so it's not he end of the world if it produces a few false
    positives/negatives.

    :param stat: the ImportStats object
    :return: True if the error on this stat is a duplicate ingestion error, False if not
    """
    return stat.error and 'this file has been ingested before' in stat.error.lower()


def get_human_duration(stat):
    """
    Get the duration on the passed ImportStats object in a sensible human readable
    format. The duration on stats objects is in seconds and therefore is great for small
    values but horrendous if it took 20 minutes. The output from this function is a
    string with either the number of seconds (to 2 decimal places), the number of
    minutes (to 0 decimal places) or the number of hours (to 0 decimal places).

    :param stat: the ImportStats object
    :return: a nicely formatted duration string
    """
    if stat.duration < 60:
        return toolkit._(f'{stat.duration:.2f} seconds')
    elif stat.duration < 60 * 60:
        return toolkit._(f'{stat.duration / 60:.0f} minutes')
    else:
        return toolkit._(f'{stat.duration / (60 * 60):.0f} hours')


def get_stat_icon(stat):
    """
    Returns the fontawesome icon class(-es) to be used for the given ImportStats object.
    The return value is based on the type of stat and other factors like whether the
    operation the stat represents is still in progress.

    :param stat: the ImportStats object
    :return: the fontawesome icon classes to use, as a string
    """
    if stat.in_progress:
        # a spinner, that spins
        return 'fa-spinner fa-pulse'
    if stat.error:
        if is_duplicate_ingestion(stat):
            # we don't want this to look like an error
            return 'fa-copy'
        return 'fa-exclamation'

    if stat.type == stats.INGEST:
        return 'fa-tasks'
    if stat.type == stats.INDEX:
        return 'fa-search'
    if stat.type == stats.PREP:
        return 'fa-cogs'
    # shouldn't get here, just use some default tick thing
    return 'fa-check'


def get_stat_activity_class(stat):
    """
    Returns the activity css class to use for the given ImportStats object. The return
    value is a string css class which either matches one of the activity item classes
    from core ckan or matches one of the custom ones in this extension's css.

    :param stat: the ImportStats object
    :return: a string
    """
    if stat.in_progress:
        return 'in_progress'
    if stat.error:
        if is_duplicate_ingestion(stat):
            return 'duplicate'
        return 'failure'
    return stat.type


def get_stat_title(stat):
    """
    Returns the title to use for the activity item created for the given ImportStats
    object. This is based on the stat's type.

    :param stat: the ImportStats object
    :return: the title for the activity item as a unicode string
    """
    if stat.type == stats.INGEST:
        return toolkit._('Ingested new resource data')
    if stat.type == stats.INDEX:
        return toolkit._('Updated search index with resource data')
    if stat.type == stats.PREP:
        return toolkit._('Validated and prepared the data for ingestion')
    return stat.type


def get_available_formats():
    """
    Simply returns all the formats that we can ingest.

    :return: a list of formats
    """
    return ALL_FORMATS


def pretty_print_json(json_string):
    """
    Does what you'd expect really.

    :param json_string: a json string
    :return: a string of pretty json
    """
    return json.dumps(json_string, sort_keys=True, indent=2)


def get_version_date(version):
    """
    Returns a date object from a version number.

    :param version: a resource/record version number (i.e. a timestamp in ms)
    :return: a date object
    """
    return date.fromtimestamp(int(version) / 1000)


def latest_item_version(resource_id, record_id=None):
    """
    Returns the most recent version for the given resource or record.

    :param resource_id: the id of the resource to search in
    :param record_id: optional; a record id to search for instead
    :return: if record id is provided, the latest record version, else the latest resource version
    """
    action = (
        'datastore_get_record_versions'
        if record_id
        else 'datastore_get_resource_versions'
    )
    data_dict = {'resource_id': resource_id}
    if record_id:
        data_dict['id'] = record_id

    versions = toolkit.get_action(action)({}, data_dict)
    return versions[-1]


def nav_slug(
    query=None, version=None, resource_ids=None, resource_ids_and_versions=None
):
    """
    Just a helper proxy for create_nav_slug.
    """
    is_new, slug = create_nav_slug(
        {}, query or {}, version, resource_ids, resource_ids_and_versions
    )
    return slug.get_slug_string()



import pytest
from ckan.model import Session
from mock import patch, MagicMock

from ckanext.versioned_datastore.lib.downloads.download import DownloadRunManager
from ckanext.versioned_datastore.logic.actions.meta.arg_objects import (
    QueryArgs,
    DerivativeArgs,
    ServerArgs,
    NotifierArgs,
)
from ckanext.versioned_datastore.model.downloads import (
    DownloadRequest,
    CoreFileRecord,
    DerivativeFileRecord,
)
from tests.helpers import patches


@pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore')
@pytest.mark.usefixtures('with_plugins', 'with_versioned_datastore_tables')
class TestDownloadRunManager:
    def test_create_run_manager(self):
        query_args = QueryArgs(query={}, query_version='v1.0.0')
        derivative_args = DerivativeArgs(format='csv')
        server_args = ServerArgs(**ServerArgs.defaults)
        notifier_args = NotifierArgs(type='none')

        # database should be empty before we start
        assert Session.query(CoreFileRecord).count() == 0
        assert Session.query(DerivativeFileRecord).count() == 0
        assert Session.query(DownloadRequest).count() == 0

        with patches.get_available_resources(), patches.rounded_versions():
            run_manager = DownloadRunManager(
                query_args, derivative_args, server_args, notifier_args
            )

        core_files = Session.query(CoreFileRecord).all()
        derivatives = Session.query(DerivativeFileRecord).all()
        request_records = Session.query(DownloadRequest).all()
        assert len(core_files) == 1
        assert len(derivatives) == 1
        assert len(request_records) == 1
        assert request_records[0].core_id == core_files[0].id
        assert request_records[0].derivative_id == derivatives[0].id
        assert derivatives[0].core_id == core_files[0].id

        assert run_manager.query.query == {}
        assert run_manager.query.query_version == 'v1.0.0'

        assert run_manager.derivative_options.format == 'csv'

        assert run_manager.notifier.name == 'none'

    def test_download_before_run(self):
        mock_plugin = MockPlugin()
        query_args = QueryArgs(query={}, query_version='v1.0.0')
        derivative_args = DerivativeArgs(format='csv')
        server_args = ServerArgs(**ServerArgs.defaults)
        notifier_args = NotifierArgs(type='none')

        with patch(
            'ckanext.versioned_datastore.lib.downloads.download.PluginImplementations',
            return_value=[mock_plugin],
        ), patches.get_available_resources(), patches.rounded_versions():
            run_manager = DownloadRunManager(
                query_args, derivative_args, server_args, notifier_args
            )

        assert run_manager.derivative_options.format == 'dwc'


class MockPlugin:
    def download_before_run(
        self, query_args, derivative_args, server_args, notifier_args
    ):
        derivative_args.format = 'dwc'
        return query_args, derivative_args, server_args, notifier_args

    def download_after_init(self, request):
        return



from ckan.model import meta, DomainObject
from ckan.model.types import JsonDictType
from sqlalchemy import Column, DateTime, Float, Boolean, Table, BigInteger, UnicodeText

# this table stores general statistics about the ingest and index events that occur on resources. It
# is also used to figure out what versions have been ingested and to a certain extent indexed and
# it's therefore pretty important (it is used to avoid ingesting older versions for example).
import_stats_table = Table(
    'versioned_datastore_import_stats',
    meta.metadata,
    Column('id', BigInteger, primary_key=True),
    Column('resource_id', UnicodeText, nullable=False, index=True),
    # the type of operation, either ingest or index
    Column('type', UnicodeText, nullable=False),
    # the version this operation is creating (for ingestion this means the version it's adding to
    # mongo, for indexing this means the version it's pulling from mongo and putting into
    # elasticsearch)
    Column('version', BigInteger, nullable=False),
    # the start datetime of the operation
    Column('start', DateTime),
    # the end datetime of the operation
    Column('end', DateTime),
    # how long the operation took in seconds
    Column('duration', Float),
    # whether the operation is in progress or whether it has completed
    Column('in_progress', Boolean),
    # if there was an error, this column is populated with the details
    Column('error', UnicodeText),
    # the number of records handled during the operation
    Column('count', BigInteger),
    # the detailed operation breakdown returned by splitgill
    Column('operations', JsonDictType),
)


class ImportStats(DomainObject):
    """
    Object for holding resource import stats.
    """

    pass


meta.mapper(ImportStats, import_stats_table)



import os
from abc import ABCMeta, abstractmethod


class BaseDerivativeGenerator(metaclass=ABCMeta):
    """
    A factory class for generating derivative files in a given format.
    """

    name = 'base'
    extension = None

    RESOURCE_ID_FIELD_NAME = 'Source resource ID'

    def __init__(self, output_dir, fields, query, resource_id=None, **format_args):
        self.output_dir = output_dir
        self.output_name = os.extsep.join(
            [resource_id or 'resource', self.extension or self.name]
        )

        if resource_id:
            self.all_fields = fields + [self.RESOURCE_ID_FIELD_NAME]
        else:
            self.all_fields = fields
        # split the fields by file; there should not be any keys here that are not in
        # self.file_paths (though this only has to contain keys for files with fields, e.g. not
        # manifest files or similar)
        self.fields = {'main': self.all_fields}

        # some derivatives might need access to the original query
        self._query = query

        self.resource_id = resource_id

        # holds any extra args not captured by the inherited __init__()
        self.format_args = format_args

        # a derivative may have multiple component files, but most will just have the one (the
        # output file). None of them _have_ to be the output filename.
        self.file_paths = {'main': os.path.join(self.output_dir, self.output_name)}
        # this will contain open file handles
        self.files = {}
        # indicators
        self._initialised = False
        self._opened = False
        self._validated = False

    def __enter__(self):
        for fn, fp in self.file_paths.items():
            self.files[fn] = open(fp, 'a')
        self._opened = True
        self.setup()
        if not self._initialised:
            self.initialise()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.finalise()
        for f in self.files.values():
            try:
                f.close()
            except AttributeError:
                pass
        self.files = {}
        self._opened = False

    @property
    def main_file(self):
        if 'main' in self.files:
            return self.files['main']
        if len(self.files.values()) == 1:
            return list(self.files.values())[0]
        else:
            return

    def initialise(self):
        """
        Runs after files have opened, before any records are processed.

        Only runs the first time the files are opened; in a multi-resource generator,
        files may be opened multiple times, but this will only be run once. Use setup()
        for things that need to be run every time.
        """
        self._initialised = True

    def setup(self):
        """
        Runs every time files are opened, before any records are processed.

        Runs before initialise().
        """
        pass

    def validate(self, record):
        """
        Runs when the first record is processed.
        """
        self._validated = True

    def write(self, record):
        if not self._validated:
            self.validate(record)
        self._write(record)

    def finalise(self):
        """
        Runs when files close.
        """
        pass

    def cleanup(self):
        """
        Runs when the generator has finished, i.e. after a single resource in separate-
        resources requests and after all resources in combined requests.
        """
        pass

    @abstractmethod
    def _write(self, record):
        raise NotImplemented



import json
from ckan.plugins import toolkit
from elasticsearch_dsl import Q

# the geo field is always meta.geo
FIELD = 'meta.geo'


def add_point_filter(search, distance, coordinates):
    """
    Adds a point filter query to the search object and returns a new search object.

    :param search: the current elasticsearch DSL object
    :param distance: the radius of the circle centred on the specified location within which records
                     must lie to be matched. This can specified in any form that elasticsearch
                     accepts for distances (see their doc, but essentially values like 10km etc).
    :param coordinates: the point to centre the radius on, specified as a lon/lat pair in a list
                        (i.e. [40.2, -20]).
    :return: a search object
    """
    options = {
        'distance': distance,
        FIELD: {
            'lat': float(coordinates[1]),
            'lon': float(coordinates[0]),
        },
    }
    return search.filter('geo_distance', **options)


def add_multipolygon_filter(search, coordinates):
    """
    Adds a multipolygon filter query to the search object and returns a new search
    object. Only the first group in each polygon grouping will be used as elasticsearch
    doesn't support this type of query yet (this kind of query is used for vacating
    space inside a polygon, like a donut for example.

    If more than one group is included then they are included as an or with a minimum must match of
    1.

    :param search: the current elasticsearch DSL object
    :param coordinates: a list of a list of a list of a list of at least 3 lon/lat pairs (i.e.
                        [[[[-16, 44], [-13.1, 34.8], [15.99, 35], [5, 49]]]])
    :return: a search object
    """
    filters = []
    for group in coordinates:
        points = group[0]
        if len(points) < 3:
            raise toolkit.ValidationError(
                'Not enough points in the polygon, must be 3 or more'
            )

        options = {
            FIELD: {
                # format the polygon point data appropriately
                'points': [
                    {'lat': float(point[1]), 'lon': float(point[0])} for point in points
                ],
            },
        }
        filters.append(Q('geo_polygon', **options))
    # add the filter to the search as an or
    return search.filter(Q('bool', should=filters, minimum_should_match=1))


def add_polygon_filter(search, coordinates):
    """
    Adds a polygon filter query to the search object and returns a new search object.
    Only the first group in each polygon grouping will be used as elasticsearch doesn't
    support this type of query yet (this kind of query is used for vacating space inside
    a polygon, like a donut for example.

    If more than one group is included then they are included as an or with a minimum must match of
    1.

    :param search: the current elasticsearch DSL object
    :param coordinates: a list of a list of a list of at least 3 lon/lat pairs (i.e. [[[-16, 44],
                        [-13.1, 34.8], [15.99, 35], [5, 49]]])
    :return: a search object
    """
    # just wrap in another list and pass to the multipolygon handler
    return add_multipolygon_filter(search, [coordinates])


def add_geo_search(search, geo_filter):
    """
    Updates the given search DSL object with the geo filter specified in the geo_filter
    dict.

    :param search: the current elasticsearch DSL object
    :param geo_filter: a dict describing a geographic filter. This should be a GeoJSON geometry and
                       should therefore include a type key and a coordinates key. The type must be
                       one of: Point, MultiPolygon or Polygon. In the case of a Point, a distance
                       key is also required which specifies the radius of the point in a form
                       elasticsearch understands (for example, 10km).
    :return: a search DSL object
    """
    # we support 3 GeoJSON types currently, Point, MultiPolygon and Polygon
    query_type_map = {
        'Point': (add_point_filter, {'distance', 'coordinates'}),
        'MultiPolygon': (add_multipolygon_filter, {'coordinates'}),
        'Polygon': (add_polygon_filter, {'coordinates'}),
    }

    try:
        # if it hasn't been parsed, parse the geo_filter as JSON
        if not isinstance(geo_filter, dict):
            geo_filter = json.loads(geo_filter)
        # fetch the function which will build the query into the search object
        add_function, required_params = query_type_map[geo_filter['type']]
    except (TypeError, ValueError):
        raise toolkit.ValidationError('Invalid geo filter information, must be JSON')
    except KeyError:
        raise toolkit.ValidationError(
            'Invalid query type, must be point, box or polygon'
        )

    try:
        # try and pull out the required parameters for each type
        parameters = {param: geo_filter[param] for param in required_params}
    except KeyError:
        required = ', '.join(required_params)
        raise toolkit.ValidationError(f'Missing parameters, must include {required}')

    # call the function which will update the search object to include the geo filters and then
    # return the resulting search object
    return add_function(search, **parameters)



from splitgill.indexing.utils import DOC_TYPE
from mock import patch, MagicMock

from ckanext.versioned_datastore.lib.common import ALL_FORMATS, DATASTORE_ONLY_RESOURCE
from ckanext.versioned_datastore.lib.datastore_utils import (
    is_datastore_resource,
    is_ingestible,
    is_datastore_only_resource,
    iter_data_fields,
)


class TestDatastoreUtils(object):
    def test_is_datastore_resource(self):
        scenarios = [
            (True, True, {'beans-banana': MagicMock()}),
            (False, False, {}),
            (False, False, {'beans-banana': MagicMock()}),
            (False, True, {}),
            (False, True, {'banana': MagicMock()}),
            (False, True, {'beans-cheese': MagicMock()}),
        ]

        def prefix_mock(name):
            return f'beans-{name}'

        for expected_outcome, exists, status in scenarios:
            client_mock = MagicMock(
                indices=MagicMock(exists=MagicMock(return_value=exists))
            )
            search_helper_mock = MagicMock(
                get_latest_index_versions=MagicMock(return_value=status)
            )

            with patch(
                'ckanext.versioned_datastore.lib.datastore_utils.prefix_resource',
                new=prefix_mock,
            ):
                with patch(
                    'ckanext.versioned_datastore.lib.common.SEARCH_HELPER',
                    new=search_helper_mock,
                ):
                    with patch(
                        'ckanext.versioned_datastore.lib.common.ES_CLIENT',
                        new=client_mock,
                    ):
                        assert expected_outcome == is_datastore_resource('banana')

    def test_is_datastore_only_resource(self):
        for yes in [
            DATASTORE_ONLY_RESOURCE,
            f'http://{DATASTORE_ONLY_RESOURCE}',
            f'https://{DATASTORE_ONLY_RESOURCE}',
        ]:
            assert is_datastore_only_resource(yes)

        for no in [
            f'ftp://{DATASTORE_ONLY_RESOURCE}',
            'this is datastore only',
            None,
            f'{DATASTORE_ONLY_RESOURCE}/{DATASTORE_ONLY_RESOURCE}',
            f'https://{DATASTORE_ONLY_RESOURCE}/nope',
        ]:
            assert not is_datastore_only_resource(no)

    def test_is_ingestible(self):
        # all formats should be ingestible (even in uppercase)
        for fmt in ALL_FORMATS:
            assert is_ingestible({'format': fmt, 'url': MagicMock()})
            assert is_ingestible({'format': fmt.upper(), 'url': MagicMock()})
        # zip should be ingestible (even in uppercase)
        assert is_ingestible({'format': 'ZIP', 'url': MagicMock()})
        assert is_ingestible({'format': 'zip', 'url': MagicMock()})
        # a datastore only resource should be ingestible
        assert is_ingestible({'format': None, 'url': DATASTORE_ONLY_RESOURCE})

        # if there's no url then the resource is not ingestible
        assert not is_ingestible({'url': None})
        assert not is_ingestible({'format': 'csv', 'url': None})
        # if there's no format and the resource is not datastore only then it is not ingestible
        assert not is_ingestible({'format': None, 'url': 'http://banana.com/test.csv'})

    def test_is_iter_data_fields(self):
        id_config = MagicMock()
        banana_config = MagicMock()
        llama_config = MagicMock()
        cheese_config = MagicMock()

        mapping = {
            'mappings': {
                DOC_TYPE: {
                    'properties': {
                        'data': {
                            'properties': {
                                '_id': id_config,
                                'nests': {
                                    'properties': {
                                        'banana': banana_config,
                                        'llama': llama_config,
                                    }
                                },
                                'cheese': cheese_config,
                            }
                        }
                    }
                }
            }
        }

        fields_and_configs = dict(iter_data_fields(mapping))
        assert fields_and_configs[('_id',)] == id_config
        assert fields_and_configs[('nests', 'banana')] == banana_config
        assert fields_and_configs[('nests', 'llama')] == llama_config
        assert fields_and_configs[('cheese',)] == cheese_config



import logging

from splitgill.ingestion.converters import RecordToMongoConverter
from splitgill.ingestion.feeders import IngestionFeeder, BaseRecord
from splitgill.ingestion.ingesters import Ingester
from splitgill.mongo import get_mongo

from .. import stats
from ... import common

log = logging.getLogger(__name__)


class DeletionRecord(BaseRecord):
    """
    A record that represents a deleted record to the converter object.
    """

    def __init__(self, version, resource_id, record_id):
        '''
        :param version: the version of this operation
        :param resource_id: the resource id of the resource we're deleting from
        :param record_id: the record to be deleted
        '''
        super(DeletionRecord, self).__init__(version)
        self.resource_id = resource_id
        self.record_id = record_id

    def convert(self):
        """
        Converts the data into the form to be stored in mongo under the "data" field. As
        this is a deletion record the return is an empty dict to signify the deletion.

        :return: an empty dict
        """
        # to signal a deletion we're using an empty dict
        return {}

    @property
    def id(self):
        """
        Returns the id of the record to delete.

        :return: the record id
        """
        return self.record_id

    @property
    def mongo_collection(self):
        """
        Returns the name of the collection in mongo that this record exists in (if
        indeed it does).

        :return: the resource id which is used as the collection name too
        """
        return self.resource_id


class DeletionFeeder(IngestionFeeder):
    """
    A feeder for deletion records.
    """

    def __init__(self, version, resource_id):
        '''
        :param version: the version of data to be fed
        :param resource_id: the resource id for which the data applies
        '''
        super(DeletionFeeder, self).__init__(version)
        self.resource_id = resource_id

    @property
    def source(self):
        """
        This is used for stats/logging and as it the name of the source of that data, in
        this case we just return "deletion" always because deletions are a special case.

        :return: "deletion", always
        """
        return 'deletion'

    def records(self):
        """
        Generator function which yields DeletionRecord objects.

        :return: yields DeletionRecords
        """
        with get_mongo(common.CONFIG, collection=self.resource_id) as mongo:
            # loop through records in mongo
            for record in mongo.find(projection=['id', 'data']):
                # only delete the record if it hasn't already been deleted
                if record['data']:
                    yield DeletionRecord(self.version, self.resource_id, record['id'])


def delete_resource_data(resource_id, version, start):
    """
    Update all the records in mongo to a deleted state (empty data). Essentially we
    present an empty dict as the new data version for all records in the collection that
    don't already have this state.

    :param resource_id: the resource id
    :param version: the version to put this ingest in as
    :param start: the start time of the deletion
    :return: True if the deletion was successful, False if not
    """
    log.info(f'Starting deletion for {resource_id}')

    # create a stats entry so that progress can be tracked
    stats_id = stats.start_operation(resource_id, stats.INGEST, version, start)
    # create a feeder object to feed the existing records into the converter
    feeder = DeletionFeeder(version, resource_id)
    # create a converter object to actually create the updates that are run on the mongo collection
    converter = RecordToMongoConverter(version, start)
    # create an ingester using our deletion feeder and the converter
    ingester = Ingester(version, feeder, converter, common.CONFIG)
    # setup monitoring on the ingester so that we can update the database with stats about the
    # ingestion as it progresses
    stats.monitor_ingestion(stats_id, ingester)

    try:
        ingester.ingest()
        return True
    except Exception as e:
        stats.mark_error(stats_id, e)
        log.exception(f'An error occurred during data deletion of {resource_id}')
        return False


class ReplaceDeletionFeeder(IngestionFeeder):
    """
    A feeder for deletion records that come from a replacement upload.
    """

    def __init__(self, version, resource_id, tracker, original_source):
        '''
        :param version: the version of data to be fed
        :param resource_id: the resource id for which the data applies
        :param tracker: the InclusionTracker object
        :param original_source: the name of the original resource data source
        '''
        super(ReplaceDeletionFeeder, self).__init__(version)
        self.resource_id = resource_id
        self.tracker = tracker
        self.original_source = original_source

    @property
    def source(self):
        return self.original_source

    def records(self):
        """
        Generator function which yields DeletionRecord objects. When the replace flag is
        true during ingestion this indicates that any records not present in the new
        resource data should be deleted. By using the tracker this feeder can yield
        record objects which represent a record that was not included in a new version
        of a resource's data for deletion.

        :return: yields DeletionRecords
        """
        with get_mongo(common.CONFIG, collection=self.resource_id) as mongo:
            # this finds all the records that haven't been updated in the given version
            for mongo_doc in mongo.find({'latest_version': {'$lt': self.version}}):
                if not self.tracker.was_included(mongo_doc['id']):
                    # delete
                    yield DeletionRecord(
                        self.version, self.resource_id, mongo_doc['id']
                    )



import pytest
from mock import MagicMock, patch
from tests.helpers import patches

from ckanext.versioned_datastore.lib.downloads.notifiers import (
    EmailNotifier,
    NullNotifier,
    WebhookNotifier,
)

notifiers = [
    (EmailNotifier, {'emails': ['data@nhm.ac.uk']}),
    (NullNotifier, {}),
    (WebhookNotifier, {'url': 'http://localhost'}),
]


@pytest.mark.parametrize('notifier_type,notifier_args', notifiers)
class TestNotifierGetText:
    def test_notifier_start_text(self, notifier_type, notifier_args):
        notifier = notifier_type(MagicMock(), **notifier_args)

        with patches.url_for():
            start_text = notifier.start_text()

        assert isinstance(start_text, tuple)
        assert len(start_text) == 2
        for t in start_text:
            assert '/banana' in t

    def test_notifier_end_text(self, notifier_type, notifier_args):
        notifier = notifier_type(MagicMock(), **notifier_args)

        with patches.url_for():
            end_text = notifier.end_text('/download-url-here')

        assert isinstance(end_text, tuple)
        assert len(end_text) == 2
        for t in end_text:
            assert '/download-url-here' in t

    def test_notifier_error_text(self, notifier_type, notifier_args):
        notifier = notifier_type(MagicMock(), **notifier_args)

        with patches.url_for():
            error_text = notifier.error_text()

        assert isinstance(error_text, tuple)
        assert len(error_text) == 2
        for t in error_text:
            assert '/banana' in t


@patch('ckanext.versioned_datastore.lib.downloads.notifiers.email.mailer')
class TestEmailNotifier:
    def test_email_notifier_start(self, mock_mailer):
        test_email_address = 'test@email.address'
        notifier = EmailNotifier(MagicMock(), emails=[test_email_address])

        with patches.url_for():
            start_text = notifier.start_text()
            notifier.notify_start()

        assert mock_mailer.mail_recipient.called
        args, kwargs = mock_mailer.mail_recipient.call_args
        assert kwargs['recipient_email'] == test_email_address
        assert kwargs['recipient_name'] == 'Downloader'
        assert kwargs['subject'] == 'Data download started'
        assert start_text[0] in kwargs['body']
        assert start_text[1] in kwargs['body_html']

    def test_email_notifier_end(self, mock_mailer):
        test_email_address = 'test@email.address'
        test_download_url = '/download-url-here'
        notifier = EmailNotifier(MagicMock(), emails=[test_email_address])

        with patches.url_for():
            end_text = notifier.end_text(test_download_url)
            notifier.notify_end(test_download_url)

        assert mock_mailer.mail_recipient.called
        args, kwargs = mock_mailer.mail_recipient.call_args
        assert kwargs['recipient_email'] == test_email_address
        assert kwargs['recipient_name'] == 'Downloader'
        assert kwargs['subject'] == 'Data download complete'
        assert end_text[0] in kwargs['body']
        assert end_text[1] in kwargs['body_html']

    def test_email_notifier_error(self, mock_mailer):
        test_email_address = 'test@email.address'
        notifier = EmailNotifier(MagicMock(), emails=[test_email_address])

        with patches.url_for():
            error_text = notifier.error_text()
            notifier.notify_error()

        assert mock_mailer.mail_recipient.called
        args, kwargs = mock_mailer.mail_recipient.call_args
        assert kwargs['recipient_email'] == test_email_address
        assert kwargs['recipient_name'] == 'Downloader'
        assert kwargs['subject'] == 'Data download failed'
        assert error_text[0] in kwargs['body']
        assert error_text[1] in kwargs['body_html']

    def test_email_notifier_multiple_emails(self, mock_mailer):
        test_email_addresses = ['one@email.address', 'two@email.address']
        notifier = EmailNotifier(MagicMock(), emails=test_email_addresses)

        with patches.url_for():
            start_text = notifier.start_text()
            notifier.notify_start()

        assert mock_mailer.mail_recipient.call_count == len(test_email_addresses)
        _, recipient_one = mock_mailer.mail_recipient.call_args_list[0]
        _, recipient_two = mock_mailer.mail_recipient.call_args_list[1]
        for k, v in recipient_one.items():
            if k == 'recipient_email':
                continue
            assert v == recipient_two[k]
        assert recipient_one['recipient_email'] == test_email_addresses[0]
        assert recipient_two['recipient_email'] == test_email_addresses[1]


@patch('ckanext.versioned_datastore.lib.downloads.notifiers.webhook.requests')
class TestWebhookNotifier:
    def test_webhook_notifier_start(self, mock_requests):
        test_webhook_url = 'http://webhook-url'
        notifier = WebhookNotifier(MagicMock(), url=test_webhook_url)

        with patches.url_for():
            start_text = notifier.start_text()
            notifier.notify_start()

        assert mock_requests.request.called
        args, kwargs = mock_requests.request.call_args
        assert args[0] == 'GET'
        assert args[1] == test_webhook_url
        assert kwargs['params']['url'] == '/banana'
        assert kwargs['params']['text'] == start_text[0]

    def test_webhook_notifier_end(self, mock_requests):
        test_webhook_url = 'http://webhook-url'
        test_download_url = '/download-url-here'
        notifier = WebhookNotifier(MagicMock(), url=test_webhook_url)

        with patches.url_for():
            end_text = notifier.end_text(test_download_url)
            notifier.notify_end(test_download_url)

        assert mock_requests.request.called
        args, kwargs = mock_requests.request.call_args
        assert args[0] == 'GET'
        assert args[1] == test_webhook_url
        assert kwargs['params']['url'] == test_download_url
        assert kwargs['params']['text'] == end_text[0]

    def test_webhook_notifier_error(self, mock_requests):
        test_webhook_url = 'http://webhook-url'
        notifier = WebhookNotifier(MagicMock(), url=test_webhook_url)

        with patches.url_for():
            error_text = notifier.error_text()
            notifier.notify_error()

        assert mock_requests.request.called
        args, kwargs = mock_requests.request.call_args
        assert args[0] == 'GET'
        assert args[1] == test_webhook_url
        assert kwargs['params']['url'] == '/banana'
        assert kwargs['params']['text'] == error_text[0]

    def test_webhook_notifier_non_default_parameters(self, mock_requests):
        test_webhook_url = 'http://webhook-url'
        notifier = WebhookNotifier(
            MagicMock(),
            url=test_webhook_url,
            url_param='something-not-url',
            text_param='anything-but-text',
            post=True,
        )

        with patches.url_for():
            start_text = notifier.start_text()
            notifier.notify_start()

        assert mock_requests.request.called
        args, kwargs = mock_requests.request.call_args
        assert args[0] == 'POST'
        assert args[1] == test_webhook_url
        assert kwargs['params']['something-not-url'] == '/banana'
        assert kwargs['params']['anything-but-text'] == start_text[0]
        assert 'url' not in kwargs['params']
        assert 'text' not in kwargs['params']


@patch('ckanext.versioned_datastore.lib.downloads.notifiers.null.logger')
class TestNullNotifier:
    def test_null_notifier_start(self, mock_logger):
        notifier = NullNotifier(MagicMock())

        with patches.url_for():
            notifier.notify_start()

        assert mock_logger.debug.called
        args, kwargs = mock_logger.debug.call_args
        assert args[0] == 'Processing started.'

    def test_null_notifier_end(self, mock_logger):
        test_download_url = '/download-url-here'
        notifier = NullNotifier(MagicMock())

        with patches.url_for():
            notifier.notify_end(test_download_url)

        assert mock_logger.debug.called
        args, kwargs = mock_logger.debug.call_args
        assert args[0] == 'Processing ended.'

    def test_null_notifier_error(self, mock_logger):
        notifier = NullNotifier(MagicMock())

        with patches.url_for():
            notifier.notify_error()

        assert mock_logger.debug.called
        args, kwargs = mock_logger.debug.call_args
        assert args[0] == 'Processing failed.'


class ContextOverridePlugin:
    def download_modify_notifier_template_context(self, request, context):
        context['status_page'] = '/boops'
        return context

    def download_modify_notifier_start_templates(self, *templates):
        return templates

    def download_modify_notifier_end_templates(self, *templates):
        return templates

    def download_modify_notifier_error_templates(self, *templates):
        return templates


class TemplateOverridePlugin:
    def download_modify_notifier_template_context(self, request, context):
        return context

    def download_modify_notifier_start_templates(self, *templates):
        return 'plain start template', 'html start template'

    def download_modify_notifier_end_templates(self, *templates):
        return 'plain end template', 'html end template'

    def download_modify_notifier_error_templates(self, *templates):
        return 'plain error template', 'html error template'


class TestNotifierTemplateOverrides:
    def test_notifier_context_overrides(self):
        mock_plugin = ContextOverridePlugin()
        notifier = NullNotifier(MagicMock())

        with patch(
            'ckanext.versioned_datastore.lib.downloads.notifiers.base.PluginImplementations',
            return_value=[mock_plugin],
        ), patch('ckan.plugins.toolkit.url_for', return_value='/banana'):
            start_text = notifier.start_text()

        assert isinstance(start_text, tuple)
        assert len(start_text) == 2
        for t in start_text:
            assert '/banana' not in t
            assert '/boops' in t

    def test_notifier_start_text_overrides(self):
        mock_plugin = TemplateOverridePlugin()
        notifier = NullNotifier(MagicMock())

        with patch(
            'ckanext.versioned_datastore.lib.downloads.notifiers.base.PluginImplementations',
            return_value=[mock_plugin],
        ), patch('ckan.plugins.toolkit.url_for', return_value='/banana'):
            start_text = notifier.start_text()

        assert isinstance(start_text, tuple)
        assert len(start_text) == 2
        assert start_text[0] == 'plain start template'
        assert start_text[1] == 'html start template'

    def test_notifier_end_text_overrides(self):
        mock_plugin = TemplateOverridePlugin()
        notifier = NullNotifier(MagicMock())

        with patch(
            'ckanext.versioned_datastore.lib.downloads.notifiers.base.PluginImplementations',
            return_value=[mock_plugin],
        ), patch('ckan.plugins.toolkit.url_for', return_value='/banana'):
            end_text = notifier.end_text('/download-url-here')

        assert isinstance(end_text, tuple)
        assert len(end_text) == 2
        assert end_text[0] == 'plain end template'
        assert end_text[1] == 'html end template'

    def test_notifier_error_text_overrides(self):
        mock_plugin = TemplateOverridePlugin()
        notifier = NullNotifier(MagicMock())

        with patch(
            'ckanext.versioned_datastore.lib.downloads.notifiers.base.PluginImplementations',
            return_value=[mock_plugin],
        ), patch('ckan.plugins.toolkit.url_for', return_value='/banana'):
            error_text = notifier.error_text()

        assert isinstance(error_text, tuple)
        assert len(error_text) == 2
        assert error_text[0] == 'plain error template'
        assert error_text[1] == 'html error template'



# this is a namespace package
try:
    import pkg_resources

    pkg_resources.declare_namespace(__name__)
except ImportError:
    import pkgutil

    __path__ = pkgutil.extend_path(__path__, __name__)



#!/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-versioned-datastore
# Created by the Natural History Museum in London, UK
from ckan.plugins import toolkit

from . import datastore, search, downloads, status

blueprints = [datastore.blueprint, search.blueprint, status.blueprint]

is_debug = toolkit.asbool(toolkit.config.get('debug', toolkit.config.get('DEBUG')))
if is_debug:
    blueprints.append(downloads.blueprint)



from elasticsearch_dsl import Search, A
from fastavro import parse_schema
from splitgill.search import create_version_query

from .query import Query
from .. import common
from ..datastore_utils import (
    prefix_resource,
    prefix_field,
    iter_data_fields,
    unprefix_index,
)
from ..query.fields import get_mappings


def get_schemas(query: Query):
    """
    Creates an avro schema from the elasticsearch index metadata.

    :param query: the Query object for this request
    :return: a parsed avro schema
    """
    # get the mappings for the resources which would have a mapping (i.e. exclude
    # non-datastore resources)
    resource_mapping = get_mappings(
        [
            resource_id
            for resource_id, version in query.resource_ids_and_versions.items()
            if version != common.NON_DATASTORE_VERSION
        ]
    )

    basic_avro_types = [
        'null',
        'boolean',
        'int',
        'long',
        'float',
        'double',
        'bytes',
        'string',
    ]
    avro_types = basic_avro_types + [
        {'type': 'array', 'items': basic_avro_types.copy()}
    ]
    avro_map_type = {'type': 'map', 'values': avro_types.copy()}
    object_avro_types = [
        'null',
        avro_map_type,
        {'type': 'array', 'items': avro_map_type.copy()},
    ]

    def _get_nest_level(field_data, nest_level=1):
        if 'properties' in field_data:
            return max(
                [
                    _get_nest_level(fd, nest_level + 1)
                    for fd in field_data['properties'].values()
                ]
            )
        if field_data['type'] == 'object':
            # 20 is the default nesting limit for ES but that makes things so slow that
            # it breaks, so we'll just hope that our data have max 5 levels
            # https://www.elastic.co/guide/en/elasticsearch/reference/master/mapping-settings-limit.html
            return 5
        else:
            return nest_level

    resource_schemas = {}
    for prefixed_resource, mapping in resource_mapping.items():
        schema_fields = {}
        field_list = mapping['mappings']['_doc']['properties']['data']['properties']
        for field_name, type_dict in field_list.items():
            max_nest_level = _get_nest_level(type_dict)
            if max_nest_level > 1:
                nested_type = object_avro_types.copy()
                for level in range(max_nest_level):
                    nested_type += [
                        {'type': 'array', 'items': nested_type.copy()},
                        {'type': 'map', 'values': nested_type.copy()},
                    ]
                schema_fields[field_name] = {'name': field_name, 'type': nested_type}
            else:
                schema_fields[field_name] = {'name': field_name, 'type': avro_types}
        schema_json = {
            'type': 'record',
            'name': 'ResourceRecord',
            'fields': list(schema_fields.values()),
        }
        resource_schemas[unprefix_index(prefixed_resource)] = parse_schema(schema_json)

    return resource_schemas


def get_fields(field_counts, ignore_empty_fields, resource_id=None):
    """
    Return a sorted list of field names for the resource ids specified using the field
    counts dict. The field counts dict should provide the field names available and
    their counts at the given version for the given search, for each resource in the
    search. If ignore_empty_fields is True, then fields with a count of 0 will be
    ignored and not returned in the list.

    The list is sorted in ascending order using lowercase comparisons.

    :param field_counts: the dict of resource ids -> fields -> counts
    :param ignore_empty_fields: whether fields with no values should be included in the
                                resulting list or not
    :param resource_id: the resource id to get the fields for. The default is None which
                        means that the fields from all resources will be returned
    :return: a list of fields in case-insensitive ascending order
    """
    # TODO: retrieve the sort order for resources from the database and use
    field_names = set()

    if resource_id:
        field_counts_list = [field_counts[resource_id]]
    else:
        field_counts_list = list(field_counts.values())

    for counts in field_counts_list:
        for field, count in counts.items():
            if count == 0 and ignore_empty_fields:
                continue
            else:
                field_names.add(field)

    return sorted(field_names, key=lambda f: f.lower())


def calculate_field_counts(query, es_client, resource_id, resource_version):
    """
    Given a download request and an elasticsearch client to work with, work out the
    number of values available per field, per resource for the search.

    :param query: the Query object
    :param es_client: the elasticsearch client to use
    :param resource_id:
    :param resource_version:
    :return: a dict of resource ids -> fields -> counts
    """
    field_counts = {}
    index_name = prefix_resource(resource_id)
    # get the base field mapping for the index so that we know which fields to look up,
    # this will get all fields from all versions and therefore isn't usable straight off
    # the bat, we have to then go and see which fields are present in the search at this
    # version
    mapping = es_client.indices.get_mapping(index_name)[index_name]

    search = (
        Search.from_dict(query.search.to_dict())
        .index(index_name)
        .using(es_client)
        .extra(size=0)
        .filter(create_version_query(resource_version))
    )

    # get all the fields names and use dot notation for nested fields
    fields = ['.'.join(parts) for parts, _config in iter_data_fields(mapping)]
    for field in fields:
        # add a search which finds the documents that have a value for the given field
        # at the right version
        agg = A('value_count', field=prefix_field(field))
        search.aggs.bucket(field, agg)

    response = search.execute()
    for field in fields:
        field_counts[field] = response.aggregations[field].value

    return field_counts


def filter_data_fields(data, field_counts, prefix=None):
    """
    Returns a new dict containing only the keys and values from the given data dict
    where the corresponding field in the field_counts dict has a value greater than 0 -
    i.e. removes all fields from the data dict that shouldn't be included.

    Note that this may seem like a pointless exercise as surely if the field count is 0
    for a field then it won't appear in any of the data dicts - however, because the
    data returned from elasticsearch is the source dict that was uploaded to it, it
    could contain nulls and empty string values. The calculate_field_counts function
    above that generates the field counts does it by using exist queries and because
    these count indexed values it skips nulls and empty strings.

    :param data: the data dict
    :param field_counts: a dict of field names and counts, the field names should be dot
                         separated for nested fields
    :param prefix: the prefix under which the fields in the passed data dict exist -
                   this is used to produce the field names for nested fields
    :return: a new dict containing only the fields from the original data dict that had
             a value other than 0 in the fields_count dict
    """
    filtered_data = {}
    for field, value in data.items():
        if prefix is not None:
            path = f'{prefix}.{field}'
        else:
            path = field

        # if the field contains a list of dicts we need to recurse for each one
        if isinstance(value, list) and value and isinstance(value[0], dict):
            filtered_value = []
            for element in value:
                filtered_element = filter_data_fields(
                    element, field_counts, prefix=path
                )
                # if there is any data left in the element after filtering, add it to
                # the temp list
                if filtered_element:
                    filtered_value.append(filtered_element)
            # if there are any dicts left from the filtering, include them directly
            # using the name of the field, not the path. We don't need to check if the
            # field has any values because we know that it does because there are dicts
            # left in the filtered list
            if filtered_value:
                filtered_data[field] = filtered_value
        # if the field is a dict, recurse to filter
        elif isinstance(value, dict):
            filtered_value = filter_data_fields(value, field_counts, prefix=path)
            # if there is any data left after the filtering, include the dict value
            # directly, using the name of the field, not the path. We don't need to
            # check if the field has any values because we know that it does because the
            # filtered_value isn't empty
            if filtered_value:
                filtered_data[field] = filtered_value
        # for everything else, just check that the path is in the fields_count dict
        elif field_counts.get(path, 0):
            filtered_data[field] = value

    return filtered_data


def flatten_dict(data, path=None, separator=' | '):
    """
    Flattens a given dictionary so that nested dicts and lists of dicts are available
    from the root of the dict. For nested dicts, the keys in the nested dict are simply
    concatenated to the key that references the dict with a dot between each, for
    example:

        {"a": {"b": 4, "c": 6}} -> {"a.b": 4, "a.c": 6}
    This works to any nesting level.
    For lists of dicts, the common keys between them are pulled up to the level above
    in the same way as the standard nested dict, but if there are multiple dicts with
    the same keys the values associated with them are concatenated together using the
    separator parameter. For example:
        {"a": [{"b": 5}, {"b": 19}]} -> {"a.b": "5 | 19"}
    :param data: the dict to flatten
    :param path: the path to place all found keys under, by default this is None and
                 therefore the keys in the dict are not placed under anything and are
                 used as is. This is really only here for internal recursive purposes.
    :param separator: the string to use when concatenating lists of values, whether
                      common ones from a list of dicts, or indeed just a normal list of
                      values
    :return: the flattened dict
    """
    flat = {}
    for key, value in data.items():
        if path is not None:
            # use a dot to indicate this key is below the parent in the path
            key = f'{path}.{key}'

        if isinstance(value, dict):
            # flatten the nested dict then update the current dict we've got on the go
            flat.update(flatten_dict(value, path=key))
        elif isinstance(value, list):
            if all(isinstance(element, dict) for element in value):
                for element in value:
                    # iterate through the list of dicts flattening each as we go and
                    # then either just adding the value to the dict we've got on the go
                    # or appending it to the string value we're using for collecting
                    # multiples
                    for subkey, subvalue in flatten_dict(element, path=key).items():
                        if subkey not in flat:
                            flat[subkey] = subvalue
                        else:
                            flat[subkey] = f'{flat[subkey]}{separator}{subvalue}'
            else:
                flat[key] = separator.join(map(str, value))
        else:
            flat[key] = value

    return flat



from traceback import format_exception_only

from mock import MagicMock, patch

from ckanext.versioned_datastore.helpers import (
    is_duplicate_ingestion,
    get_human_duration,
    get_stat_icon,
    get_stat_activity_class,
    get_stat_title,
    get_available_formats,
)
from ckanext.versioned_datastore.lib.common import ALL_FORMATS
from ckanext.versioned_datastore.lib.importing.ingestion.exceptions import (
    DuplicateDataSource,
    UnsupportedDataSource,
)
from ckanext.versioned_datastore.lib.importing.stats import (
    INGEST,
    INDEX,
    PREP,
    ALL_TYPES,
)


class TestHelpers(object):
    def test_is_duplicate_ingestion(self):
        dup_exception = DuplicateDataSource('some_file_hash')

        # should be able to match on just the message
        stat1 = MagicMock(error=str(dup_exception))
        assert is_duplicate_ingestion(stat1)

        # but also the final line of the actual stack output
        stat2 = MagicMock(
            error=str(
                format_exception_only(DuplicateDataSource, dup_exception)[-1].strip()
            )
        )
        assert is_duplicate_ingestion(stat2)

        # it shouldn't match other things, for example a UnsupportedDataSource exception
        non_dup_exception = UnsupportedDataSource('csv')

        # just the message should fail
        stat3 = MagicMock(error=str(non_dup_exception))
        assert not is_duplicate_ingestion(stat3)

        # and so should the final line of the actual stack output
        stat4 = MagicMock(
            error=str(
                format_exception_only(UnsupportedDataSource, non_dup_exception)[
                    -1
                ].strip()
            )
        )
        assert not is_duplicate_ingestion(stat4)

    def test_get_human_duration(self):
        scenarios = [
            # seconds
            (10.381, '10.38 seconds'),
            (10, '10.00 seconds'),
            (0, '0.00 seconds'),
            (2.111111111111328042384, '2.11 seconds'),
            (2.3, '2.30 seconds'),
            (10.385, '10.38 seconds'),
            # minutes
            (60, '1 minutes'),
            (61, '1 minutes'),
            (190, '3 minutes'),
            (280.30290, '5 minutes'),
            # hours
            (3600, '1 hours'),
            (3600.3289, '1 hours'),
            (11900, '3 hours'),
            (60 * 60 * 24 * 365 * 3, '26280 hours'),
        ]
        for duration, expected_output in scenarios:
            stat = MagicMock(duration=duration)
            assert get_human_duration(stat) == expected_output

    def test_get_stat_icon(self):
        # in progress stats are always pulsing spinners regardless of their type
        assert get_stat_icon(MagicMock(in_progress=True)) == 'fa-spinner fa-pulse'

        # if there's a non-duplicate error, we use exclamation
        with patch(
            'ckanext.versioned_datastore.helpers.is_duplicate_ingestion',
            MagicMock(return_value=False),
        ):
            assert (
                get_stat_icon(MagicMock(in_progress=False, error=MagicMock()))
                == 'fa-exclamation'
            )

        # if there's a duplicate error, we use copy
        with patch(
            'ckanext.versioned_datastore.helpers.is_duplicate_ingestion',
            MagicMock(return_value=True),
        ):
            assert (
                get_stat_icon(MagicMock(in_progress=False, error=MagicMock()))
                == 'fa-copy'
            )

        # now check the types
        assert (
            get_stat_icon(MagicMock(in_progress=False, error=None, type=INGEST))
            == 'fa-tasks'
        )
        assert (
            get_stat_icon(MagicMock(in_progress=False, error=None, type=INDEX))
            == 'fa-search'
        )
        assert (
            get_stat_icon(MagicMock(in_progress=False, error=None, type=PREP))
            == 'fa-cogs'
        )

        # check that no types are missing icons
        for stat_type in ALL_TYPES:
            assert (
                get_stat_icon(MagicMock(in_progress=False, error=None, type=stat_type))
                != 'fa-check'
            )

        # anything else gets a check to avoid erroring
        assert (
            get_stat_icon(MagicMock(in_progress=False, error=None, type='banana'))
            == 'fa-check'
        )

    def test_get_stat_activity_class(self):
        # in progress stats always get the in_progress class
        assert get_stat_activity_class(MagicMock(in_progress=True)) == 'in_progress'

        # if there's a non-duplicate error, we use failure
        with patch(
            'ckanext.versioned_datastore.helpers.is_duplicate_ingestion',
            MagicMock(return_value=False),
        ):
            assert (
                get_stat_activity_class(MagicMock(in_progress=False, error=MagicMock()))
                == 'failure'
            )

        # if there's a duplicate error, we use duplicate
        with patch(
            'ckanext.versioned_datastore.helpers.is_duplicate_ingestion',
            MagicMock(return_value=True),
        ):
            assert (
                get_stat_activity_class(MagicMock(in_progress=False, error=MagicMock()))
                == 'duplicate'
            )

        # now check the types. For these we just return the actual type value as the return value
        for stat_type in ALL_TYPES + [MagicMock()]:
            assert (
                get_stat_activity_class(
                    MagicMock(in_progress=False, error=None, type=stat_type)
                )
                == stat_type
            )

    def test_get_stat_title(self):
        # just check that for the types we know about we don't get back the default
        for stat_type in ALL_TYPES:
            assert get_stat_title(MagicMock(type=stat_type)) != stat_type

        fake_type = MagicMock()
        assert get_stat_title(MagicMock(type=fake_type)) == fake_type

    def test_get_available_formats(self):
        formats = get_available_formats()
        assert isinstance(formats, list)
        for f in ALL_FORMATS:
            assert f in formats



from ckan.plugins import toolkit
from flask import Blueprint, send_from_directory

blueprint = Blueprint(name='downloads', import_name=__name__, url_prefix='/downloads')


@blueprint.route('/direct/<zip_name>')
def direct(zip_name):
    """
    Serves up the requested zip from the download directory. This is only registered
    with flask when running in debug mode and therefore is only for testing in
    development. In production we should always serve files through the web server.

    :param zip_name: the zip name
    :return: the send file response
    """
    download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
    return send_from_directory(download_dir, zip_name)


@blueprint.route('/custom/<zip_name>')
def custom(zip_name):
    """
    Basically exactly the same as the /direct/ route, but for serving the symlinks with
    custom filenames.

    :param zip_name: the zip name
    :return: the send file response
    """
    download_dir = toolkit.config.get('ckanext.versioned_datastore.custom_dir')
    return send_from_directory(download_dir, zip_name)



import csv
import json
import os
import pytest
import shutil
import tempfile
import zipfile
from mock import patch, MagicMock
from tests.helpers import patches, data as test_data

from ckan.plugins import toolkit
from ckan.tests import factories
from ckanext.versioned_datastore.model.downloads import DownloadRequest

scenarios = [
    ('csv', {}),
    ('json', {}),
    ('xlsx', {}),
    ('dwc', {}),
    ('csv', {'delimiter': 'tab'}),
    (
        'dwc',
        {
            'core_extension_name': 'gbif_occurrence',
            'extension_names': ['gbif_multimedia'],
            'extension_map': {'gbif_multimedia': ['img']},
        },
    ),
]

expected_extensions = {'csv': '.csv', 'json': '.json', 'xlsx': '.xlsx', 'dwc': '.zip'}


@pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore')
@pytest.mark.usefixtures(
    'with_plugins',
    'with_versioned_datastore_tables',
    'with_vds_resource',
    'clear_download_dir',
)
class TestQueueDownload:
    @patches.enqueue_job()
    @pytest.mark.parametrize('file_format,format_args', scenarios)
    @pytest.mark.parametrize('separate_files', [True, False])
    def test_run_download_formats(
        self, enqueue_job, with_vds_resource, file_format, format_args, separate_files
    ):
        with patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {'query': {}},
                    'file': {
                        'format': file_format,
                        'format_args': format_args,
                        'separate_files': separate_files,
                    },
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            if not separate_files:
                assert f'resource{expected_extensions[file_format]}' in archive_files
                assert len(archive_files) == 2
            else:
                assert len(archive_files) == 3
                assert (
                    f'{with_vds_resource[0]["id"]}{expected_extensions[file_format]}'
                    in archive_files
                )
                assert (
                    f'{with_vds_resource[1]["id"]}{expected_extensions[file_format]}'
                    in archive_files
                )

    @patches.enqueue_job()
    def test_run_download_without_query(self, enqueue_job):
        with patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {'query': {}},
                    'file': {'format': 'csv'},
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        self.temp_dir = tempfile.mktemp()
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            zf.extract('manifest.json', self.temp_dir)
            assert 'resource.csv' in archive_files
            zf.extract('resource.csv', self.temp_dir)

        with open(os.path.join(self.temp_dir, 'manifest.json')) as f:
            manifest = json.load(f)
            assert manifest['download_id'] == download_details['download_id']
            assert manifest['file_format'] == 'csv'
            assert sorted(manifest['files']) == sorted(archive_files)
            assert not manifest['ignore_empty_fields']
            assert not manifest['separate_files']
            assert manifest['total_records'] == len(
                test_data.records + test_data.records_addtl
            )

        with open(os.path.join(self.temp_dir, 'resource.csv')) as f:
            reader = csv.DictReader(f)
            records = [row for row in reader]
            assert len(records) == len(test_data.records + test_data.records_addtl)
            for record in records:
                assert 'emptyField' in record
                test_data_record = next(
                    r
                    for r in test_data.records + test_data.records_addtl
                    if r['scientificName'] == record['scientificName']
                )
                for k, v in test_data_record.items():
                    assert v == record[k]

        shutil.rmtree(self.temp_dir)

    @patches.enqueue_job()
    def test_run_download_with_query(self, enqueue_job, with_vds_resource):
        with patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {
                        'query': {
                            'filters': {
                                'and': [
                                    {
                                        'string_equals': {
                                            'fields': ['colour'],
                                            'value': 'green',
                                        }
                                    }
                                ]
                            },
                        }
                    },
                    'file': {'format': 'csv'},
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        self.temp_dir = tempfile.mktemp()
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            assert 'resource.csv' in archive_files
            zf.extract('resource.csv', self.temp_dir)

        with open(os.path.join(self.temp_dir, 'resource.csv')) as f:
            reader = csv.DictReader(f)
            records = [row for row in reader]
            assert len(records) == 1  # the number of records that should match the q

        shutil.rmtree(self.temp_dir)

    @patches.enqueue_job()
    def test_run_download_keep_empty(self, enqueue_job, with_vds_resource):
        with patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {
                        'query': {
                            'filters': {
                                'and': [
                                    {
                                        'string_equals': {
                                            'fields': ['group'],
                                            'value': 'b',
                                        }
                                    }
                                ]
                            },
                        }
                    },
                    'file': {
                        'format': 'csv',
                        'ignore_empty_fields': False,
                        'separate_files': False,
                    },
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        self.temp_dir = tempfile.mktemp()
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            assert 'resource.csv' in archive_files
            zf.extract('resource.csv', self.temp_dir)

        with open(os.path.join(self.temp_dir, 'resource.csv')) as f:
            reader = csv.reader(f)
            header = [row for row in reader][0]
            assert 'emptyField' in header

        shutil.rmtree(self.temp_dir)

    @patches.enqueue_job()
    def test_run_download_ignore_empty(self, enqueue_job, with_vds_resource):
        with patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {
                        'query': {
                            'filters': {
                                'and': [
                                    {
                                        'string_equals': {
                                            'fields': ['group'],
                                            'value': 'b',
                                        }
                                    }
                                ]
                            },
                        }
                    },
                    'file': {
                        'format': 'csv',
                        'ignore_empty_fields': True,
                        'separate_files': False,
                    },
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        self.temp_dir = tempfile.mktemp()
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            assert 'resource.csv' in archive_files
            zf.extract('resource.csv', self.temp_dir)

        with open(os.path.join(self.temp_dir, 'resource.csv')) as f:
            reader = csv.reader(f)
            header = [row for row in reader][0]
            assert 'emptyField' not in header

        shutil.rmtree(self.temp_dir)

    @patches.enqueue_job()
    @pytest.mark.parametrize('transform', [{'id_as_url': {'field': 'urlSlug'}}])
    def test_run_download_with_transform(self, enqueue_job, transform):
        with patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {'query': {}},
                    'file': {'format': 'csv', 'transform': transform},
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        self.temp_dir = tempfile.mktemp()
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            assert 'resource.csv' in archive_files
            zf.extract('resource.csv', self.temp_dir)

        with open(os.path.join(self.temp_dir, 'resource.csv')) as f:
            reader = csv.DictReader(f)
            records = [row for row in reader]
            assert len(records) == len(test_data.records + test_data.records_addtl)
            for record in records:
                assert record['urlSlug'].endswith('/banana')

        shutil.rmtree(self.temp_dir)

    @patches.enqueue_job()
    def test_run_download_with_non_vds_resource(self, enqueue_job, with_vds_resource):
        non_ds_resource = factories.Resource(url_type='upload')

        def _shutil_mock(src, dest):
            # there's nothing to copy from, so just write the new file
            with open(dest, 'w') as f:
                f.write('hello')

        with patch('shutil.copy2', side_effect=_shutil_mock), patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {
                        'query': {},
                        'resource_ids': [non_ds_resource['id']],
                    },
                    'file': {
                        'format': 'raw',
                        'format_args': {'allow_non_datastore': True},
                    },
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        self.temp_dir = tempfile.mktemp()
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            zf.extract('manifest.json', self.temp_dir)
            assert len(archive_files) == 2
            # it's easier to do it this way than to predict the url or split it or whatever
            extless_files = [os.path.splitext(f)[0] for f in archive_files]
            assert non_ds_resource['id'] in extless_files

        with open(os.path.join(self.temp_dir, 'manifest.json')) as f:
            manifest = json.load(f)
            assert manifest['download_id'] == download_details['download_id']
            assert manifest['file_format'] == 'raw'
            assert sorted(manifest['files']) == sorted(archive_files)
            assert not manifest['ignore_empty_fields']
            assert manifest['separate_files']
            assert manifest['total_records'] == 1

        shutil.rmtree(self.temp_dir)


@pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore query_dois')
@pytest.mark.ckan_config('ckanext.query_dois.prefix', 'xx.xxxx')
@pytest.mark.usefixtures(
    'with_plugins',
    'with_versioned_datastore_tables',
    'with_vds_resource',
    'clear_download_dir',
)
class TestDownloadWithQueryDois:
    @classmethod
    def setup_class(cls):
        from ckanext.query_dois.model import query_doi_table, query_doi_stat_table

        cls.tables = [query_doi_table, query_doi_stat_table]

        for table in cls.tables:
            table.create(checkfirst=True)

    @classmethod
    def teardown_class(cls):
        for table in cls.tables:
            if table.exists():
                table.drop()

    @patches.enqueue_job()
    def test_run_download_with_query_dois(self, enqueue_job):
        # I cannot get the query_doi endpoints to load for the life of me so we're just
        # going to mock it before I lose my mind
        def _url_for_mock(endpoint, **kwargs):
            if endpoint == 'query_doi.landing_page':
                return f'/{kwargs["data_centre"]}/{kwargs["identifier"]}'
            else:
                return '/banana'

        with patch(
            'ckanext.query_dois.lib.doi.find_existing_doi',
            return_value=MagicMock(doi='123456'),
        ), patch('ckan.plugins.toolkit.url_for', _url_for_mock):
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {'query': {}},
                    'file': {'format': 'dwc'},
                    'notifier': {'type': 'none'},
                },
            )
        enqueue_job.assert_called()
        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        assert any(
            [
                f.startswith(download_request.derivative_record.download_hash)
                for f in os.listdir(download_dir)
            ]
        )


@pytest.mark.ckan_config('ckan.plugins', 'versioned_datastore')
@pytest.mark.usefixtures(
    'with_plugins',
    'with_versioned_datastore_tables',
    'with_vds_resource',
    'clear_download_dir',
)
@patches.enqueue_job()
class TestDownloadInterfaces:
    def test_modify_args(self, enqueue_job):
        mock_plugin = ModifyArgsPlugin()

        with patch(
            'ckanext.versioned_datastore.lib.downloads.download.PluginImplementations',
            return_value=[mock_plugin],
        ), patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {'query': {}},
                    'file': {'format': 'csv'},
                    'notifier': {'type': 'none'},
                },
            )

        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            assert 'resource.xlsx' in archive_files
            assert len(archive_files) == 2

    def test_modify_manifest(self, enqueue_job):
        mock_plugin = ModifyManifestPlugin()

        with patch(
            'ckanext.versioned_datastore.lib.downloads.download.PluginImplementations',
            return_value=[mock_plugin],
        ), patches.url_for():
            download_details = toolkit.get_action('datastore_queue_download')(
                {},
                {
                    'query': {'query': {}},
                    'file': {'format': 'csv'},
                    'notifier': {'type': 'none'},
                },
            )

        download_request = DownloadRequest.get(download_details['download_id'])
        assert download_request is not None
        download_dir = toolkit.config.get('ckanext.versioned_datastore.download_dir')
        matching_zips = [
            f
            for f in os.listdir(download_dir)
            if f.startswith(download_request.derivative_record.download_hash)
        ]
        assert len(matching_zips) == 1
        temp_dir = tempfile.mktemp()
        with zipfile.ZipFile(os.path.join(download_dir, matching_zips[0]), 'r') as zf:
            archive_files = zf.namelist()
            assert 'manifest.json' in archive_files
            assert 'resource.csv' in archive_files
            assert len(archive_files) == 2
            zf.extract('manifest.json', temp_dir)

        with open(os.path.join(temp_dir, 'manifest.json')) as f:
            manifest = json.load(f)
            assert manifest['download_id'] == download_details['download_id']
            assert manifest['totally-new-key'] == 'bananas'

        shutil.rmtree(temp_dir)


class ModifyArgsPlugin:
    def download_before_run(
        self, query_args, derivative_args, server_args, notifier_args
    ):
        derivative_args.format = 'xlsx'
        return query_args, derivative_args, server_args, notifier_args

    def download_modify_manifest(self, manifest, request):
        return manifest

    def download_after_run(self, request):
        return

    def download_after_init(self, request):
        return


class ModifyManifestPlugin:
    def download_before_run(self, *args):
        return args

    def download_modify_manifest(self, manifest, request):
        manifest['totally-new-key'] = 'bananas'
        return manifest

    def download_after_run(self, request):
        return

    def download_after_init(self, request):
        return



import io
import json
from collections import OrderedDict

import abc
import itertools
import os
import six
from jsonschema.validators import validator_for, RefResolver
from importlib_resources import files

schemas = OrderedDict()
schema_base_path = files('ckanext.versioned_datastore.theme').joinpath(
    'public/querySchemas'
)


def register_schema(version, schema):
    """
    Registers a new schema with the given version into the central schemas dict. The
    schema parameter should be a subclass of the Schema class but generally must at
    least provide the translate and validate methods. After registration, the schemas
    dict is updated to ensure the correct sort order is in use.

    :param version: the query schema version
    :param schema: the Schema object representing the query schema
    """
    global schemas
    # add the new version and re-sort the schemas by version in ascending order
    # (h/t https://stackoverflow.com/a/2574090)
    schemas = OrderedDict(
        sorted(
            itertools.chain([(version, schema)], schemas.items()),
            key=lambda vs: [int(u) for u in vs[0][1:].split('.')],
        )
    )


def get_latest_query_version():
    """
    Gets the latest query version from the registered schemas dict.

    :return: the latest query version
    """
    return next(iter(schemas.keys()))


class InvalidQuerySchemaVersionError(Exception):
    def __init__(self, version):
        super(Exception, self).__init__(f'Invalid query version: {version}')


def validate_query(query, version):
    """
    Validate the given query dict against the query schema for the given version. If the
    version doesn't match any registered schemas then an InvalidQuerySchemaVersionError
    will be raised.

    :param query: the query dict
    :param version: the query schema version to validate against
    :return: True if the validation succeeded, otherwise jsonschema exceptions will be raised
    """
    if version not in schemas:
        raise InvalidQuerySchemaVersionError(version)
    schemas[version].validate(query)
    return True


def translate_query(query, version, search=None):
    """
    Translates the given query dict into an elasticsearch-dsl object using the Schema
    object associated with the given version. If the version doesn't match any
    registered schemas then an InvalidQuerySchemaVersionError will be raised.

    :param query: the whole query dict
    :param version: the query schema version to translate using
    :param search: an instantiated elasticsearch-dsl object to be built on instead of creating
                   a fresh object. By default a new search object is created.
    :return: an instantiated elasticsearch-dsl object
    """
    if version not in schemas:
        raise InvalidQuerySchemaVersionError(version)
    else:
        return schemas[version].translate(query, search=search)


def hash_query(query, version):
    """
    Hashes the given query at the given version and returns the unique digest.

    :param query: the query dict
    :param version: the query version
    :return: the hash
    """
    if version not in schemas:
        raise InvalidQuerySchemaVersionError(version)
    else:
        return schemas[version].hash(query)


def load_core_schema(version):
    """
    Given a query schema version, loads the schema from the schema_base_path directory.

    :param version: the version to load
    :return: the loaded schema (as a dict) and a jsonschmea validator object for the schema
    """
    schema_file = schema_base_path.joinpath(version).joinpath(f'{version}.json')
    with io.open(schema_file, 'r', encoding='utf-8') as f:
        schema = json.load(f)
        validator_cls = validator_for(schema)
        validator_cls.check_schema(schema)
        # create a resolver which can resolve refs relative to the schema
        resolver = RefResolver(base_uri=f'file://{schema_file}', referrer=schema)
        validator = validator_cls(schema, resolver=resolver)
        return schema, validator


@six.add_metaclass(abc.ABCMeta)
class Schema(object):
    """
    Abstract base class for a query schema.

    Should be extended by all registered query schemas.
    """

    @abc.abstractmethod
    def validate(self, query):
        """
        Validate the given query against this schema. Failures are marked raising
        jsonschema exceptions.

        :param query: the query dict to validate
        """
        pass

    @abc.abstractmethod
    def translate(self, query, search=None):
        """
        Translates the query into an elasticsearch-dsl search object.

        :param query: the whole query dict
        :param search: an instantiated elasticsearch-dsl object to be built on instead of creating
                       a fresh object. By default a new search object is created.
        :return: an instantiated elasticsearch-dsl object
        """
        pass

    @abc.abstractmethod
    def hash(self, query):
        """
        Hashes the query and returns the hex digest.

        :param query: the whole query dict
        :return: a string hex digest
        """
        pass



from datetime import datetime

from ckan.plugins import toolkit
from splitgill.search import create_version_query
from splitgill.utils import to_timestamp, chunk_iterator
from elasticsearch_dsl import Search, MultiSearch

from .meta import help, schema
from ckantools.decorators import action
from ...lib import common
from ...lib.basic_query.search import create_search
from ...lib.datastore_utils import (
    prefix_resource,
    is_datastore_resource,
    get_public_alias_name,
)
from ...lib.query.schema import get_latest_query_version


@action(
    schema.datastore_get_record_versions(),
    help.datastore_get_record_versions,
    toolkit.side_effect_free,
)
def datastore_get_record_versions(id, resource_id):
    """
    Retrieves all the versions of the record with the given id in the given resource.
    The versions are returned as a list in ascending order.

    :param id: the id of the record
    :param resource_id: the id of the resource the record is in
    :return: a list of versions in ascending order
    """
    index_name = prefix_resource(resource_id)
    return common.SEARCH_HELPER.get_record_versions(index_name, int(id))


@action(
    schema.datastore_search(),
    help.datastore_get_resource_versions,
    toolkit.side_effect_free,
)
def datastore_get_resource_versions(
    resource_id, context, data_dict, original_data_dict
):
    """
    Retrieves all the versions of the given resource when under the given search. Note
    that the schema used for this action is the same as the datastore_search schema. The
    return is a list of dicts each of which includes the version timestamp, the number
    of records modified in the version and the total records at the version.

    :param resource_id: the id of the resource to examine
    :param context: the context dict from the action call
    :param data_dict: the data_dict from the action call
    :param original_data_dict: the data_dict before it was validated
    :return: a list of dicts
    """
    original_data_dict, data_dict, version, search = create_search(
        context, data_dict, original_data_dict
    )
    index_name = prefix_resource(resource_id)

    # this gives us every version in the index, plus the number of changes in that
    # version (changes include new records and changed records)
    counts = common.SEARCH_HELPER.get_index_version_counts(index_name, search=search)

    # each of the dicts in the counts list above contains the version and the number of
    # changes, but not the number of records in that version, we want to add this to
    # the dicts. To do this we will run msearches against elasticsearch so that we can
    # batch up the searches and get better performance (there could be 1000s of
    # versions to count after all). This variable simply defines how many searches to do
    # in each msearch batch
    multisearch_chunk_size = 100

    for details_chunk in chunk_iterator(counts, multisearch_chunk_size):
        multisearch = MultiSearch(using=common.ES_CLIENT, index=index_name)
        for details in details_chunk:
            multisearch = multisearch.add(
                Search()[0:0].filter(create_version_query(details["version"]))
            )
        results = multisearch.execute()
        # update the count details we got from splitgill with the actual record count
        for detail, result in zip(details_chunk, results):
            detail["count"] = result.hits.total

    return counts


@action(
    schema.datastore_get_rounded_version(),
    help.datastore_get_rounded_version,
    toolkit.side_effect_free,
)
def datastore_get_rounded_version(resource_id, version=None):
    """
    Retrieve the closest version of the resource's data to the given version when
    rounding down.

    :param resource_id: the id of the resource
    :param version: the version timestamp. If None (the default) the latest version of the resource
                    is returned
    :return: the rounded version timestamp
    """
    index_name = prefix_resource(resource_id)
    return common.SEARCH_HELPER.get_rounded_versions([index_name], version)[index_name]


@action(
    schema.datastore_is_datastore_resource(),
    help.datastore_is_datastore_resource,
    toolkit.side_effect_free,
)
def datastore_is_datastore_resource(resource_id):
    """
    Checks whether the given resource id is in the datastore or not.

    :param resource_id: the resource to check
    :return: True if it is, False if not
    """
    return is_datastore_resource(resource_id)


@action({}, help.datastore_get_latest_query_schema_version, get=True)
def datastore_get_latest_query_schema_version():
    """
    Simply returns the latest available query schema version.

    :return: the query schema version
    """
    return get_latest_query_version()


@action(schema.datastore_count(), help.datastore_count, get=True)
def datastore_count(resource_ids=None, version=None):
    if version is None:
        version = to_timestamp(datetime.now())
    if resource_ids is None:
        resource_ids = ['*']

    indexes = [get_public_alias_name(resource_id) for resource_id in resource_ids]
    search = Search(using=common.ES_CLIENT, index=indexes).filter(
        create_version_query(version)
    )

    return search.count()



import base64
import json
import zlib

from ckan.model import Session
from ckan.plugins import toolkit
from flask import Blueprint
from sqlalchemy import func

from ..lib.query import slugs
from ..lib.query.slug_words import list_one, list_two, list_three
from ..model.slugs import DatastoreSlug

blueprint = Blueprint(name='search', import_name=__name__)


@blueprint.route('/search')
@blueprint.route('/search/<path:slug>')
def view(slug=''):
    return toolkit.render('search/search.html', extra_vars={'slug': slug})


@blueprint.route('/slugerator')
def slug_generator():
    """
    A queryless slug generator.

    Because why not.
    :return:
    """
    extra_vars = {}
    slug_q = Session.query(DatastoreSlug)

    total_permutations = len(list_one) * len(list_two) * len(list_three)
    extra_vars['total'] = total_permutations

    pretty_slugs = slug_q.filter(DatastoreSlug.pretty_slug.isnot(None)).count()
    extra_vars['unreserved'] = pretty_slugs
    extra_vars['remaining'] = total_permutations - pretty_slugs
    extra_vars['percent_used'] = round((pretty_slugs / total_permutations) * 100, 2)

    reserved_slugs = slug_q.filter(
        DatastoreSlug.reserved_pretty_slug.isnot(None)
    ).count()
    extra_vars['reserved'] = reserved_slugs

    date_func = func.date_trunc('day', DatastoreSlug.created)
    q = Session.query(date_func.label('date'), func.count().label('count'))
    q = q.order_by(date_func).group_by(date_func)
    slug_stats = []
    total_slugs = 0
    for stat in q.all():
        total_slugs += stat.count
        formatted_date = stat.date.strftime('%Y-%m-%d')
        slug_stats.append([formatted_date, total_slugs])
    extra_vars['slug_stats'] = slug_stats
    # also compress the data and add to toolkit.c
    toolkit.c.date_interval = 'day'
    toolkit.c.graph_data = base64.b64encode(
        zlib.compress(json.dumps(slug_stats).encode(), level=9)
    )

    attempts = 0
    while attempts < 100:
        new_slug = slugs.generate_pretty_slug()
        if slugs.resolve_slug(new_slug) is None:
            extra_vars['new_slug'] = new_slug
        attempts += 1
        new_slug = None  # reset if it didn't work
    # default to words not on any list if not already set
    extra_vars['new_slug'] = extra_vars.get('new_slug', 'something-something-something')

    return toolkit.render('search/slugerator.html', extra_vars=extra_vars)



from datetime import datetime

from sqlalchemy import (
    Column,
    Table,
    BigInteger,
    UnicodeText,
    DateTime,
    ForeignKey,
    desc,
)
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import relationship, backref
from sqlalchemy.exc import InvalidRequestError

from ckan.model import meta, DomainObject, Session
from ckan.model.types import make_uuid

# this one is outside DownloadRequest so we can use it as a default in the table def
state_initial = 'initiated'

# describes the core files from which the output derivatives (e.g. CSV, DwC-A) are
# generated
datastore_downloads_core_files_table = Table(
    'vds_download_core',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column('query_hash', UnicodeText, nullable=False, index=True),
    Column('query', JSONB, nullable=False),
    Column('query_version', UnicodeText, nullable=False),
    Column('resource_ids_and_versions', JSONB, nullable=False, default=dict),
    Column('resource_hash', UnicodeText, nullable=False),
    Column('modified', DateTime, nullable=False, default=datetime.utcnow),
    Column('total', BigInteger, nullable=True),
    Column('resource_totals', JSONB, nullable=False, default=dict),
    Column('field_counts', JSONB, nullable=False, default=dict),
)

# describes derived files generated from the core files
datastore_downloads_derivative_files_table = Table(
    'vds_download_derivative',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column(
        'core_id',
        UnicodeText,
        ForeignKey('vds_download_core.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=False,
    ),
    Column('download_hash', UnicodeText, nullable=False, index=True),
    Column('created', DateTime, nullable=False, default=datetime.utcnow),
    Column('format', UnicodeText, nullable=False),
    Column('options', JSONB, nullable=True),
    Column('filepath', UnicodeText, nullable=True),
)

datastore_downloads_requests_table = Table(
    'vds_download_request',
    meta.metadata,
    Column('id', UnicodeText, primary_key=True, default=make_uuid),
    Column('created', DateTime, nullable=False, default=datetime.utcnow),
    Column('modified', DateTime, nullable=False, default=datetime.utcnow),
    Column('state', UnicodeText, nullable=False, default=state_initial),
    Column('message', UnicodeText, nullable=True),
    Column('server_args', JSONB, nullable=True),
    Column(
        'derivative_id',
        UnicodeText,
        ForeignKey(
            'vds_download_derivative.id', onupdate='CASCADE', ondelete='CASCADE'
        ),
        nullable=True,
    ),
    Column(
        'core_id',
        UnicodeText,
        ForeignKey('vds_download_core.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=True,
    ),
)


class CoreFileRecord(DomainObject):
    """
    A record of a generated core download file.

    Does not track individual downloads.
    """

    id: str
    query_hash: str
    query: dict
    query_version: str
    resource_ids_and_versions: dict
    resource_hash: str
    modified: datetime
    total: int
    resource_totals: dict
    field_counts: dict
    derivatives: list
    requests: list

    @classmethod
    def get(cls, record_id):
        return Session.query(cls).get(record_id)

    def update(self, **kwargs):
        for field, value in kwargs.items():
            setattr(self, field, value)
        self.modified = datetime.utcnow()
        try:
            self.save()
        except InvalidRequestError:
            self.commit()

    @classmethod
    def get_by_hash(cls, query_hash, resource_hash):
        return (
            Session.query(cls)
            .filter(cls.query_hash == query_hash, cls.resource_hash == resource_hash)
            .order_by(desc(cls.modified))
            .all()
        )

    @classmethod
    def find_resource(cls, query_hash, resource_id, resource_version, exclude=None):
        exclude = exclude or []
        have_resource = (
            Session.query(cls)
            .filter(
                cls.query_hash == query_hash,
                cls.resource_ids_and_versions.has_key(resource_id),
                cls.resource_totals.has_key(resource_id),
                cls.id.notin_(exclude),
            )
            .order_by(desc(cls.modified))
            .all()
        )
        have_resource_version = [
            r
            for r in have_resource
            if r.resource_ids_and_versions[resource_id] == resource_version
        ]
        if have_resource_version:
            return have_resource_version[0]
        return


class DerivativeFileRecord(DomainObject):
    """
    A record of a download file requested by a user in a specific format.

    Does not track individual downloads, just lists the options used to generate a file.
    """

    id: str
    core_id: str
    download_hash: str
    created: datetime
    format: str
    options: dict
    filepath: str
    core_record: CoreFileRecord
    requests: list

    @classmethod
    def get(cls, record_id):
        return Session.query(cls).get(record_id)

    def update(self, **kwargs):
        for field, value in kwargs.items():
            setattr(self, field, value)
        try:
            self.save()
        except InvalidRequestError:
            self.commit()

    @classmethod
    def get_by_hash(cls, download_hash):
        return (
            Session.query(cls)
            .filter(cls.download_hash == download_hash)
            .order_by(desc(cls.created))
            .all()
        )

    @classmethod
    def get_by_filepath(cls, filepath):
        return (
            Session.query(cls)
            .filter(cls.filepath == filepath)
            .order_by(desc(cls.created))
            .all()
        )


class DownloadRequest(DomainObject):
    """
    A record of an individual download request.
    """

    id: str
    created: datetime
    modified: datetime
    state: str
    message: str
    server_args: dict
    derivative_id: str
    derivative_record: DerivativeFileRecord
    core_record: CoreFileRecord

    state_initial = state_initial
    state_complete = 'complete'
    state_failed = 'failed'
    state_packaging = 'zipping'
    state_derivative_gen = 'gen_derivative'
    state_core_gen = 'gen_core'
    state_retrieving = 'retrieving'

    @classmethod
    def get(cls, request_id):
        return Session.query(cls).get(request_id)

    def update(self, **kwargs):
        for field, value in kwargs.items():
            setattr(self, field, value)
        self.modified = datetime.utcnow()
        try:
            self.save()
        except InvalidRequestError:
            self.commit()

    def update_status(self, status_text, message=None):
        self.update(state=status_text, message=message)


meta.mapper(CoreFileRecord, datastore_downloads_core_files_table, properties={})

meta.mapper(
    DerivativeFileRecord,
    datastore_downloads_derivative_files_table,
    properties={
        'core_record': relationship(
            CoreFileRecord,
            primaryjoin=datastore_downloads_derivative_files_table.c.core_id
            == CoreFileRecord.id,
            backref=backref('derivatives', cascade='all, delete-orphan'),
            lazy='joined',
        )
    },
)

meta.mapper(
    DownloadRequest,
    datastore_downloads_requests_table,
    properties={
        'derivative_record': relationship(
            DerivativeFileRecord,
            primaryjoin=datastore_downloads_requests_table.c.derivative_id
            == DerivativeFileRecord.id,
            backref=backref('requests', cascade='all, delete-orphan'),
            lazy='joined',
        ),
        'core_record': relationship(
            CoreFileRecord,
            primaryjoin=datastore_downloads_requests_table.c.core_id
            == CoreFileRecord.id,
            backref=backref('requests', cascade='all, delete-orphan'),
            lazy='joined',
        ),
    },
)



# !/usr/bin/env python
# encoding: utf-8
#
# This file is part of ckanext-versioned-datastore
# Created by the Natural History Museum in London, UK

from ckan.plugins import toolkit
from flask import Blueprint

from ..lib.importing.stats import get_all_stats

blueprint = Blueprint(name='datastore', import_name=__name__, url_prefix='')


@blueprint.route(
    '/dataset/<package_name>/resource_data/<resource_id>', methods=['GET', 'POST']
)
def resource_data(package_name, resource_id):
    """
    Produces the DataStore page on a resource.

    This page contains details of the resource's ingestion and indexing.
    """
    try:
        # first, check access
        toolkit.check_access('resource_update', {}, {'id': resource_id})
        # then retrieve the package and resource data
        toolkit.c.pkg_dict = toolkit.get_action('package_show')(
            {}, {'id': package_name}
        )
        toolkit.c.resource = toolkit.get_action('resource_show')(
            {}, {'id': resource_id}
        )
    except toolkit.ObjectNotFound:
        toolkit.abort(404, toolkit._('Resource not found'))
    except toolkit.NotAuthorized:
        toolkit.abort(401, toolkit._('Unauthorized to edit this resource'))

    if toolkit.request.method == 'POST':
        toolkit.get_action('datastore_reindex')({}, {'resource_id': resource_id})
        toolkit.h.flash_success(
            toolkit._(
                'Reindexing submitted, this may take a few minutes. Yo'
                'can monitor progress below'
            )
        )
        # redirect the user back to the page. This ensures they can do things like reload
        # without getting a pesky "would you like to resubmit this form" notice
        return toolkit.redirect_to(
            'datastore.resource_data',
            package_name=package_name,
            resource_id=resource_id,
            _code=303,
        )
    else:
        extra_vars = {
            'stats': get_all_stats(resource_id),
            'reindex_action': toolkit.url_for(
                'datastore.resource_data',
                package_name=package_name,
                resource_id=resource_id,
            ),
            'pkg_dict': toolkit.c.pkg_dict,
            'resource': toolkit.c.resource,
        }
        return toolkit.render('package/resource_data.html', extra_vars=extra_vars)



import logging
from contextlib import suppress
from datetime import datetime

from ckan import model
from ckan.model import DomainObjectOperation
from ckan.plugins import (
    toolkit,
    interfaces,
    SingletonPlugin,
    implements,
    PluginImplementations,
)
from splitgill.utils import to_timestamp

from . import routes, helpers, cli
from .interfaces import IVersionedDatastoreQuerySchema, IVersionedDatastore
from .lib.common import setup
from .lib.datastore_utils import (
    is_datastore_resource,
    ReadOnlyResourceException,
    InvalidVersionException,
    update_resources_privacy,
)
from .lib.query.schema import register_schema
from .lib.query.v1_0_0 import v1_0_0Schema
from .logic import auth
from .logic.actions import basic_search, crud, downloads, extras, multisearch
from ckantools.loaders import create_actions, create_auth

log = logging.getLogger(__name__)


class VersionedSearchPlugin(SingletonPlugin):
    implements(interfaces.IActions)
    implements(interfaces.IAuthFunctions)
    implements(interfaces.ITemplateHelpers, inherit=True)
    implements(interfaces.IResourceController, inherit=True)
    implements(interfaces.IDomainObjectModification, inherit=True)
    implements(interfaces.IConfigurer)
    implements(interfaces.IConfigurable)
    implements(interfaces.IBlueprint, inherit=True)
    implements(IVersionedDatastoreQuerySchema)
    implements(interfaces.IClick)

    # IActions
    def get_actions(self):
        return create_actions(basic_search, crud, downloads, extras, multisearch)

    # IClick
    def get_commands(self):
        return cli.get_commands()

    # IAuthFunctions
    def get_auth_functions(self):
        return create_auth(auth)

    # ITemplateHelpers
    def get_helpers(self):
        return {
            'is_datastore_resource': is_datastore_resource,
            'is_duplicate_ingestion': helpers.is_duplicate_ingestion,
            'get_human_duration': helpers.get_human_duration,
            'get_stat_icon': helpers.get_stat_icon,
            'get_stat_activity_class': helpers.get_stat_activity_class,
            'get_stat_title': helpers.get_stat_title,
            'get_available_formats': helpers.get_available_formats,
            'pretty_print_json': helpers.pretty_print_json,
            'get_version_date': helpers.get_version_date,
            'latest_item_version': helpers.latest_item_version,
            'nav_slug': helpers.nav_slug,
        }

    # IResourceController
    def before_show(self, resource_dict):
        resource_dict['datastore_active'] = is_datastore_resource(resource_dict['id'])
        return resource_dict

    # IDomainObjectModification
    def notify(self, entity, operation):
        """
        Respond to changes to model objects. We use this hook to ensure any new data is
        imported into the versioned datastore and to make sure the privacy settings on
        the data are up to date. We're only interested in:

            - resource deletions
            - new resources
            - resources that have had changes to their URL
            - packages that have changed

        :param entity: the entity that has changed
        :param operation: the operation undertaken on the object. This will be one of the options
                          from the DomainObjectOperation enum.
        """
        if (
            isinstance(entity, model.Package)
            and operation == DomainObjectOperation.changed
        ):
            # if a package is the target entity and it's been changed ensure the privacy is applied
            # correctly to its resource indexes
            update_resources_privacy(entity)
        elif isinstance(entity, model.Resource):
            context = {'model': model, 'ignore_auth': True}
            data_dict = {'resource_id': entity.id}

            if operation == DomainObjectOperation.deleted:
                toolkit.get_action('datastore_delete')(context, data_dict)
            else:
                do_upsert = False

                if operation == DomainObjectOperation.new:
                    # datastore_create returns True when the resource looks like it's ingestible
                    do_upsert = toolkit.get_action('datastore_create')(
                        context, data_dict
                    )
                elif operation == DomainObjectOperation.changed:
                    # always try the upsert if the resource has changed
                    do_upsert = True

                if do_upsert:
                    # in theory, last_modified should change when the resource file/url is changed
                    # and metadata_modified should change when any other attributes are changed. To
                    # cover off the possibility that this gets mixed up, we'll pick the max of them
                    modified = list(
                        filter(None, (entity.last_modified, entity.metadata_modified))
                    )
                    last_modified = max(modified) if modified else datetime.now()
                    data_dict['version'] = to_timestamp(last_modified)
                    # use replace to overwrite the existing data (this is what users would expect)
                    data_dict['replace'] = True
                    # these exceptions are fine to swallow
                    with suppress(ReadOnlyResourceException, InvalidVersionException):
                        toolkit.get_action('datastore_upsert')(context, data_dict)

    # IConfigurer
    def update_config(self, config):
        # add public folder containing schemas
        toolkit.add_public_directory(config, 'theme/public')
        # add templates
        toolkit.add_template_directory(config, 'theme/templates')
        toolkit.add_resource('theme/assets', 'ckanext-versioned-datastore')

    # IBlueprint
    def get_blueprint(self):
        return routes.blueprints

    # IConfigurable
    def configure(self, ckan_config):
        setup(ckan_config)

        # register all custom query schemas
        for plugin in PluginImplementations(IVersionedDatastoreQuerySchema):
            for version, schema in plugin.get_query_schemas():
                register_schema(version, schema)

        # reserve any requested slugs
        from .lib.query.slugs import reserve_slug

        for plugin in PluginImplementations(IVersionedDatastore):
            for (
                reserved_pretty_slug,
                query_parameters,
            ) in plugin.datastore_reserve_slugs().items():
                with suppress(Exception):
                    reserve_slug(reserved_pretty_slug, **query_parameters)

    # IVersionedDatastoreQuerySchema
    def get_query_schemas(self):
        return [(v1_0_0Schema.version, v1_0_0Schema())]



# !/usr/bin/env python
# encoding: utf-8
import io
import json

import jsonschema
import os

import pytest
from ckanext.versioned_datastore.lib.query.schema import schema_base_path
from ckanext.versioned_datastore.lib.query.v1_0_0 import v1_0_0Schema


class TestV1_0_0Translator(object):
    def test_validate_examples(self):
        schema = v1_0_0Schema()
        path = os.path.join(schema_base_path, schema.version, 'examples')
        for filename in os.listdir(path):
            with io.open(os.path.join(path, filename), 'r', encoding='utf-8') as f:
                schema.validate(json.load(f))

    @staticmethod
    def compare_query_and_search(query, search_dict):
        schema = v1_0_0Schema()
        # validate the query!
        schema.validate(query)
        # check that the translated version is correct compared to the expected search dict
        assert schema.translate(query).to_dict() == search_dict

    def test_translate_1(self):
        query = {}
        search_dict = {}
        self.compare_query_and_search(query, search_dict)

    def test_translate_2(self):
        query = {'search': 'mollusca'}
        search_dict = {
            'query': {'match': {'meta.all': {'query': 'mollusca', 'operator': 'and'}}}
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_3(self):
        query = {
            u"filters": {
                u"and": [
                    {u"string_equals": {u"fields": [u"genus"], u"value": u"helix"}},
                    {
                        u"string_contains": {
                            u"fields": [u"higherGeography"],
                            u"value": u"europe",
                        }
                    },
                ]
            }
        }
        search_dict = {
            'query': {
                'bool': {
                    'filter': [
                        {'term': {'data.genus': 'helix'}},
                        {
                            'match': {
                                'data.higherGeography.full': {
                                    'query': 'europe',
                                    'operator': 'and',
                                }
                            }
                        },
                    ]
                }
            }
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_4(self):
        query = {
            'search': 'italy',
            u"filters": {
                u"and": [
                    {u"string_equals": {u"fields": [u"genus"], u"value": u"helix"}},
                    {
                        u"string_contains": {
                            u"fields": [u"higherGeography"],
                            u"value": u"europe",
                        }
                    },
                ]
            },
        }
        search_dict = {
            'query': {
                'bool': {
                    'filter': [
                        {'term': {'data.genus': 'helix'}},
                        {
                            'match': {
                                'data.higherGeography.full': {
                                    'query': 'europe',
                                    'operator': 'and',
                                }
                            }
                        },
                    ],
                    'must': [
                        {'match': {'meta.all': {'query': 'italy', 'operator': 'and'}}}
                    ],
                }
            }
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_5(self):
        query = {
            u"filters": {
                u"and": [
                    {u"string_equals": {u"fields": [u"genus"], u"value": u"helix"}},
                    {
                        u"or": [
                            {
                                u"string_contains": {
                                    u"fields": [u"higherGeography"],
                                    u"value": u"italy",
                                }
                            },
                            {
                                u"string_contains": {
                                    u"fields": [u"higherGeography"],
                                    u"value": u"spain",
                                }
                            },
                            {
                                u"string_contains": {
                                    u"fields": [u"higherGeography"],
                                    u"value": u"portugal",
                                }
                            },
                        ]
                    },
                ]
            }
        }
        search_dict = {
            'query': {
                'bool': {
                    'filter': [
                        {'term': {'data.genus': 'helix'}},
                        {
                            'bool': {
                                'minimum_should_match': 1,
                                'should': [
                                    {
                                        'match': {
                                            'data.higherGeography.full': {
                                                'query': 'italy',
                                                'operator': 'and',
                                            }
                                        }
                                    },
                                    {
                                        'match': {
                                            'data.higherGeography.full': {
                                                'query': 'spain',
                                                'operator': 'and',
                                            }
                                        }
                                    },
                                    {
                                        'match': {
                                            'data.higherGeography.full': {
                                                'query': 'portugal',
                                                'operator': 'and',
                                            }
                                        }
                                    },
                                ],
                            }
                        },
                    ]
                }
            }
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_6(self):
        query = {
            u"filters": {
                u"and": [
                    {u"string_equals": {u"fields": [u"genus"], u"value": u"helix"}},
                    {
                        u"number_range": {
                            u"fields": [u"year"],
                            u"less_than": 2010,
                            u"less_than_inclusive": True,
                            u"greater_than": 2000,
                            u"greater_than_inclusive": True,
                        }
                    },
                    {
                        u"or": [
                            {
                                u"string_contains": {
                                    u"fields": [u"higherGeography"],
                                    u"value": u"italy",
                                }
                            },
                            {
                                u"string_contains": {
                                    u"fields": [u"higherGeography"],
                                    u"value": u"spain",
                                }
                            },
                            {
                                u"string_contains": {
                                    u"fields": [u"higherGeography"],
                                    u"value": u"portugal",
                                }
                            },
                        ]
                    },
                ]
            }
        }
        search_dict = {
            'query': {
                'bool': {
                    'filter': [
                        {'term': {'data.genus': 'helix'}},
                        {'range': {'data.year.number': {'gte': 2000, 'lte': 2010}}},
                        {
                            'bool': {
                                'minimum_should_match': 1,
                                'should': [
                                    {
                                        'match': {
                                            'data.higherGeography.full': {
                                                'query': 'italy',
                                                'operator': 'and',
                                            }
                                        }
                                    },
                                    {
                                        'match': {
                                            'data.higherGeography.full': {
                                                'query': 'spain',
                                                'operator': 'and',
                                            }
                                        }
                                    },
                                    {
                                        'match': {
                                            'data.higherGeography.full': {
                                                'query': 'portugal',
                                                'operator': 'and',
                                            }
                                        }
                                    },
                                ],
                            }
                        },
                    ]
                }
            }
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_7(self):
        query = {
            u"filters": {
                u"and": [
                    {u"string_equals": {u"fields": [u"genus"], u"value": u"helix"}},
                    {
                        u"geo_point": {
                            u"latitude": 51.4712,
                            u"longitude": -0.9421,
                            u"radius": 10,
                            u"radius_unit": u"mi",
                        }
                    },
                ]
            }
        }
        search_dict = {
            'query': {
                'bool': {
                    'filter': [
                        {'term': {'data.genus': 'helix'}},
                        {
                            'geo_distance': {
                                'meta.geo': {'lat': 51.4712, 'lon': -0.9421},
                                'distance': '10mi',
                            }
                        },
                    ]
                }
            }
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_8(self):
        query = {u"filters": {u"and": [{u"exists": {u"fields": [u"associatedMedia"]}}]}}
        search_dict = {'query': {'exists': {'field': 'data.associatedMedia'}}}
        self.compare_query_and_search(query, search_dict)

    def test_translate_9(self):
        query = {u"filters": {u"and": [{u"exists": {u"geo_field": True}}]}}
        search_dict = {'query': {'exists': {'field': 'meta.geo'}}}
        self.compare_query_and_search(query, search_dict)

    def test_translate_10(self):
        country = 'Curaçao'
        multipolygon = v1_0_0Schema().geojson['country'][country]
        query = {u"filters": {u"and": [{u"geo_named_area": {u"country": country}}]}}
        search_dict = {
            'query': {
                'geo_polygon': {
                    'meta.geo': {
                        'points': [
                            {'lat': point[1], 'lon': point[0]}
                            for point in multipolygon[0][0]
                        ]
                    }
                }
            }
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_11(self):
        a_square = [
            [102.0, 2.0],
            [103.0, 2.0],
            [103.0, 3.0],
            [102.0, 3.0],
            [102.0, 2.0],
        ]
        another_square = [
            [100.0, 0.0],
            [101.0, 0.0],
            [101.0, 1.0],
            [100.0, 1.0],
            [100.0, 0.0],
        ]
        a_hole = [[100.2, 0.2], [100.8, 0.2], [100.8, 0.8], [100.2, 0.8], [100.2, 0.2]]

        def to_points(points):
            return [{'lat': point[1], 'lon': point[0]} for point in points]

        query = {
            u"filters": {
                u"and": [
                    {
                        u"geo_custom_area": [
                            # just a square
                            [a_square],
                            # a square with a square hole in it
                            [another_square, a_hole],
                        ]
                    }
                ]
            }
        }
        search_dict = {
            u"query": {
                u"bool": {
                    u"minimum_should_match": 1,
                    u"should": [
                        {
                            u"geo_polygon": {
                                u"meta.geo": {u"points": to_points(a_square)}
                            }
                        },
                        {
                            u"bool": {
                                u"filter": [
                                    {
                                        u"geo_polygon": {
                                            u"meta.geo": {
                                                u"points": to_points(another_square)
                                            }
                                        }
                                    }
                                ],
                                u"must_not": [
                                    {
                                        u"geo_polygon": {
                                            u"meta.geo": {u"points": to_points(a_hole)}
                                        }
                                    }
                                ],
                            }
                        },
                    ],
                }
            }
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_12(self):
        query = {
            u"filters": {
                u"not": [
                    {u"string_equals": {u"fields": [u"genus"], u"value": u"helix"}}
                ]
            }
        }
        search_dict = {
            'query': {'bool': {'must_not': [{'term': {'data.genus': 'helix'}}]}}
        }
        self.compare_query_and_search(query, search_dict)

    def test_translate_ignores_additional_properties_in_filters(self):
        schema = v1_0_0Schema()

        nope = {
            u"filters": {
                'something_else': {},
                u"not": [
                    {u"string_equals": {u"fields": [u"genus"], u"value": u"helix"}}
                ],
            }
        }
        with pytest.raises(jsonschema.ValidationError):
            schema.validate(nope)

    def test_translate_ignores_additional_properties_in_geo_named_area(self):
        schema = v1_0_0Schema()

        nope = {
            u"filters": {
                u"and": [
                    {
                        u"geo_named_area": {
                            u"country": u"Belgium",
                            u"something_else": False,
                        }
                    }
                ]
            }
        }
        with pytest.raises(jsonschema.ValidationError):
            schema.validate(nope)



"""
Add new download tables.

Revision ID: 19a61e5b669f
Revises:
Create Date: 2023-01-06 10:27:56.739905
"""
import hashlib
import json
from datetime import datetime as dt
from uuid import uuid4

import sqlalchemy as sa
from alembic import op
from sqlalchemy import orm
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.ext.declarative import declarative_base

# revision identifiers, used by Alembic.
revision = '19a61e5b669f'
down_revision = None
branch_labels = None
depends_on = None

Base = declarative_base()


def make_uuid():
    return str(uuid4())


def make_hash(content):
    if isinstance(content, list):
        hashable = '|'.join(map(str, content))
    else:
        hashable = str(content)
    return hashlib.sha1(hashable.encode('utf-8')).hexdigest()


class Core(Base):
    __tablename__ = 'vds_download_core'

    id = sa.Column(sa.UnicodeText, primary_key=True, default=make_uuid)
    query_hash = sa.Column(sa.UnicodeText, nullable=False, index=True)
    query = sa.Column(JSONB, nullable=False)
    query_version = sa.Column(sa.UnicodeText, nullable=False)
    resource_ids_and_versions = sa.Column(JSONB, nullable=False, default=dict)
    resource_hash = sa.Column(sa.UnicodeText, nullable=False)
    modified = sa.Column(sa.DateTime, nullable=False, default=dt.utcnow)
    total = sa.Column(sa.BigInteger, nullable=True)
    resource_totals = sa.Column(JSONB, nullable=False, default=dict)
    field_counts = sa.Column(JSONB, nullable=False, default=dict)


class Derivative(Base):
    __tablename__ = 'vds_download_derivative'

    id = sa.Column(sa.UnicodeText, primary_key=True, default=make_uuid)
    core_id = sa.Column(
        sa.UnicodeText,
        sa.ForeignKey('vds_download_core.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=False,
    )
    download_hash = sa.Column(sa.UnicodeText, nullable=False, index=True)
    created = sa.Column(sa.DateTime, nullable=False, default=dt.utcnow)
    format = sa.Column(sa.UnicodeText, nullable=False)
    options = sa.Column(JSONB, nullable=True)
    filepath = sa.Column(sa.UnicodeText, nullable=True)

    core_record = orm.relationship('Core', backref='derivatives')


class Request(Base):
    __tablename__ = 'vds_download_request'

    id = sa.Column(sa.UnicodeText, primary_key=True, default=make_uuid)
    created = sa.Column(sa.DateTime, nullable=False, default=dt.utcnow)
    modified = sa.Column(sa.DateTime, nullable=False, default=dt.utcnow)
    state = sa.Column(sa.UnicodeText, nullable=False, default='initiated')
    message = sa.Column(sa.UnicodeText, nullable=True)
    derivative_id = sa.Column(
        sa.UnicodeText,
        sa.ForeignKey(
            'vds_download_derivative.id', onupdate='CASCADE', ondelete='CASCADE'
        ),
        nullable=True,
    )
    core_id = sa.Column(
        sa.UnicodeText,
        sa.ForeignKey('vds_download_core.id', onupdate='CASCADE', ondelete='CASCADE'),
        nullable=True,
    )

    core_record = orm.relationship('Core', backref='requests')
    derivative_record = orm.relationship('Derivative', backref='requests')


class DatastoreDownload(Base):
    __tablename__ = 'versioned_datastore_downloads'

    id = sa.Column(sa.UnicodeText, primary_key=True, default=make_uuid)
    query_hash = sa.Column(sa.UnicodeText, nullable=False, index=True)
    query = sa.Column(JSONB, nullable=False)
    query_version = sa.Column(sa.UnicodeText, nullable=False)
    resource_ids_and_versions = sa.Column(JSONB, nullable=False)
    created = sa.Column(sa.DateTime, nullable=False, default=dt.utcnow)
    total = sa.Column(sa.BigInteger, nullable=True)
    resource_totals = sa.Column(JSONB, nullable=True)
    state = sa.Column(sa.UnicodeText, nullable=True)
    error = sa.Column(sa.UnicodeText, nullable=True)
    options = sa.Column(JSONB, nullable=True)


def upgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)

    Core.__table__.create(bind)
    Derivative.__table__.create(bind)
    Request.__table__.create(bind)

    for download in session.query(DatastoreDownload):
        # calculate hashes first
        resource_hash = make_hash(sorted(download.resource_ids_and_versions.items()))
        record_hash = make_hash([download.query_hash, resource_hash])
        file_options = {  # use defaults from old datastore_queue_download def
            'format': download.options.get('format', 'csv'),
            'format_args': download.options.get('format_args', {}),
            'separate_files': download.options.get('separate_files', True),
            'ignore_empty_fields': download.options.get('ignore_empty_fields', True),
            'transform': download.options.get('transform'),
        }
        derivative_hash = make_hash(json.dumps(file_options))
        download_hash = make_hash([record_hash, derivative_hash])

        core_record = (
            session.query(Core)
            .filter(
                Core.query_hash == download.query_hash,
                Core.resource_hash == resource_hash,
            )
            .first()
        )
        if core_record:
            if (
                isinstance(download.resource_totals, dict)
                and core_record.resource_totals != download.resource_totals
            ):
                core_record.resource_totals = download.resource_totals
                core_record.modified = download.created
            if download.total is not None and core_record.total is None:
                core_record.total = download.total
                core_record.modified = download.created
        else:
            core_record = Core(
                query_hash=download.query_hash,
                query=download.query,
                query_version=download.query_version,
                resource_ids_and_versions=download.resource_ids_and_versions,
                resource_hash=resource_hash,
                modified=download.created,
                total=download.total,
                resource_totals=download.resource_totals,
            )
        session.add(core_record)
        session.commit()

        derivative_record = (
            session.query(Derivative)
            .filter(
                Derivative.download_hash == download_hash,
                Derivative.core_id == core_record.id,
            )
            .first()
        )
        if not derivative_record:
            file_format = file_options.pop('format')
            derivative_record = Derivative(
                core_id=core_record.id,
                download_hash=download_hash,
                created=download.created,
                format=file_format,
                options=file_options,
            )
            session.add(derivative_record)
            session.commit()

        new_request = Request(
            created=download.created,
            modified=download.created,
            state=download.state,
            message=download.error,
            derivative_id=derivative_record.id,
            core_id=core_record.id,
        )
        session.add(new_request)
        session.commit()

    DatastoreDownload.__table__.drop(bind)


def downgrade():
    bind = op.get_bind()
    session = orm.Session(bind=bind)

    DatastoreDownload.__table__.create(bind)

    for request in session.query(Request):
        # there are new/different states in the new model
        state = (
            'processing'
            if request.state not in ['complete', 'failed', 'zipping']
            else request.state
        )
        # broken requests may not have a linked derivative record
        if not request.derivative_id:
            options = {}
        else:
            options = request.derivative_record.options
            options['format'] = request.derivative_record.format
        # or core record
        if request.core_record:
            core_record = request.core_record
        elif request.derivative_record and request.derivative_record.core_record:
            # somehow the derivative record is linked and has a core record (this
            # shouldn't happen but you never know)
            core_record = request.derivative_record.core_record
        else:
            # this will have no info so we just have to skip it
            print(f'Skipping {request.id} (state: {request.state}) due to lack of data')
            continue

        new_download = DatastoreDownload(
            query_hash=core_record.query_hash,
            query=core_record.query,
            query_version=core_record.query_version,
            resource_ids_and_versions=core_record.resource_ids_and_versions,
            created=request.created,
            total=core_record.total,
            resource_totals=core_record.resource_totals,
            state=state,
            error=request.message if request.state == 'failed' else None,
            options=options,
        )
        session.add(new_download)
        session.commit()

    Request.__table__.drop(bind)
    Derivative.__table__.drop(bind)
    Core.__table__.drop(bind)



from .csv import CsvDerivativeGenerator
from .json import JsonDerivativeGenerator
from .xlsx import XlsxDerivativeGenerator
from .dwc import DwcDerivativeGenerator

derivatives = [
    CsvDerivativeGenerator,
    JsonDerivativeGenerator,
    XlsxDerivativeGenerator,
    DwcDerivativeGenerator,
]
